"""
Copyright 2024 - The Minton Group at Purdue University
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

import numpy as np

def wrap_angle(angle):
    """
    Converts angles to be between 0 and 360 degrees.
        
    Parameters
    ----------
    angle : float
        Angle to be converted
    Returns
    -------
    angle : float
        Converted angle
    """
    while np.any(angle >= 360.0 ):
        angle[angle >= 360.0] -= 360.0
    while np.any(angle < 0.0):
        angle[angle < 0.0] += 360.0
    return angle


def follow_swift(ds, ifol=None, nskp=None):
    """
    Emulates the Swift version of tool_follow.f
    It will generate a file called follow.out containing 10 columns (angles are all in degrees)::
    
        1 2 3  4    5     6    7    8    9   10
        t,a,e,inc,capom,omega,capm,peri,apo,obar
        
    Parameters
    ----------
    ds : Xarray Dataset 
        Dataset containing orbital elements
    ifol : int, optional
        Particle number to follow. The default is None, in which case the user is prompted to enter a particle number.
    nskp : int, optional
        Print frequency. The default is None, in which case the user is prompted to enter a print frequency. 

    Returns
    -------
    fol : Xarray Dataset
        Dataset containing only the followed body with angles converted to degrees

    """
    fol = None
    if ifol is None:
        intxt = input(' Input the particle number to follow \n')
        ifol = int(intxt)
    print(f"Following particle {ifol}")
    if ifol < 0: # Negative numbers are planets
        fol = ds.where(np.invert(np.isnan(ds['Gmass'])), drop=True)
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True) # Remove times where this body doesn't exist (but this also gets rid of the central body)
        fol = fol.isel(id = -ifol - 2)  # Take 1 off for 0-indexed arrays in Python, and take 1 more off because the central body is gone
    elif ifol > 0: # Positive numbers are test particles
        fol = ds.where(np.isnan(ds['Gmass']), drop=True).drop_vars(['Gmass', 'radius'],errors="ignore")
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True)
        fol = fol.isel(id = ifol - 1)  # Take 1 off for 0-indexed arrays in Python

    if nskp is None:
        intxt = input('Input the print frequency\n')
        nskp = int(intxt)
        
    fol['obar'] = fol['capom'] + fol['omega']
    fol['obar'] = fol['obar'].fillna(0)
    fol['obar'] = wrap_angle(fol['obar'])
    fol['peri'] = fol['a'] * (1.0 - fol['e'])
    fol['apo']  = fol['a'] * (1.0 + fol['e'])

    
    tslice = slice(None, None, nskp)
    try:
        with open('follow.out', 'w') as f:
            print('# 1 2 3  4    5     6    7    8    9   10', file=f)
            print('# t,a,e,inc,capom,omega,capm,peri,apo,obar', file=f)
            for t in fol.isel(time=tslice).time:
                a = fol['a'].sel(time=t).values
                e = fol['e'].sel(time=t).values
                inc = fol['inc'].sel(time=t).values
                capom = fol['capom'].sel(time=t).values
                omega = fol['omega'].sel(time=t).values
                capm = fol['capm'].sel(time=t).values
                peri = fol['peri'].sel(time=t).values
                apo = fol['apo'].sel(time=t).values
                obar = fol['obar'].sel(time=t).values
                print(f"{t.values:15.7e} {a:22.16f} {e:22.16f} {inc:22.16f} {capom:22.16f} {omega:22.16f} {capm:22.16f} {peri:22.16f} {apo:22.16f} {obar:22.16f}", file=f)
                
    except IOError:
        print(f"Error writing to follow.out")
    
    return fol



"""
Copyright 2024 - The Minton Group at Purdue University
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""
from __future__ import annotations

import swiftest
from . import constants
import numpy as np
from astroquery.jplhorizons import Horizons, HorizonsClass
import astropy.units as u
from astropy.coordinates import SkyCoord
import datetime
from typing import Any, Union
import warnings

def get_solar_system_body_mass_rotation(id: str,
                                        jpl: HorizonsClass=None,
                                        ephemerides_start_date: str=constants.MINTON_BCL,
                                        verbose: bool=True,
                                        **kwargs: Any) -> dict:
    """
    Parses the raw output from JPL Horizons in order to extract physical properties of a body if they exist
    

    Parameters
    ----------
    id : string or list of strings
        A string identifying which body is requested from JPL/Horizons (or a list of strings if multiple ids are possible, such as an altid list)
    jpl : HorizonsClass
        An astroquery.jplhorizons HorizonsClass object. If None, a new query will be made.
    ephemerides_start_date : string
        Date to use when obtaining the ephemerides in the format YYYY-MM-DD. Default is constants.MINTON_BCL
    verbose : bool
        Indicates whether to print messages about the query or not. Default is False
    **kwargs: Any
        Additional keyword arguments to pass to the query method (see https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html)

    Returns
    -------
    A dictionary containing the following elements
    
    Gmass : float
        G*Mass of the body in m^3/s^2
    radius : float
        The radius of the body in m
    rot : (3) float vector
        The rotation rate vector oriented toward the north pole in deg/s
    """

    def get_Gmass(raw_response):
        GM = [s for s in raw_response.split('\n') if 'GM' in s 
            and 'Rel' not in s 
            and 'GMT' not in s
            and 'ANGMOM' not in s]
        if len(GM) == 0:
            # Try an alternative name for the Mass found in some satellite queries
            M = [s for s in raw_response.split('\n') if 'Mass' in s]
            if len(M) > 0:
                M = M[0].split('Mass')[-1].strip()
                if 'kg' in M:
                    try:
                        unit_conv_str = M.split('kg')[0].strip()
                        unit_conv_str = unit_conv_str.split('^')[1].strip()
                        unit_conv = 10**int(unit_conv_str)
                        M = M.split('=')[1].strip()
                        if '^' in M:
                            mult = M.split('^')[1].split(')')[0].strip()
                            mult = 10**int(mult)
                        else:
                            mult = 1.0
                        M = M.split()[0].strip()
                        M = float(M)  * mult * unit_conv
                        return M * swiftest.GC * 1e-9 # Return units of km**3 / s**2 for consistency
                    except:
                        return None
            return None
        GM = GM[0]
        if len(GM) > 1:
            GM = GM.split('GM')[1]
            if len(GM) > 1:
                GM = GM.split('=')
                if len(GM) > 1:
                    GM = GM[1].strip().split(' ')[0].split('+')[0]
        try:
            GM = float(GM)
        except:
            GM = None
        return GM

    def get_radius(raw_response):
        radius = [s for s in raw_response.split('\n') if 'mean radius' in s.lower() or 'RAD' in s or 'Radius (km' in s]
        if len(radius) == 0:
            return None
        radius = radius[0]
        if "Radius (km" in radius or "RAD" in radius:
            if "Radius (km" in radius: 
                radius = radius.split("Radius (km")[1].strip().strip('=').split('+')[0].strip()
            elif "RAD" in radius:
                radius = radius.split('RAD')[1].strip().strip('=').split('+')[0].strip().split()[0].strip()
            if "=" in radius:
                radius = radius.split('=')[1].strip()
            if 'x' in radius: # Triaxial ellipsoid bodies like Haumea may have multiple dimensions which need to be averaged
                radius = radius.split('x')
                try:
                    for i,v in enumerate(radius):
                        radius[i] = float(v.split()[0].strip())
                    radius = np.average(radius)
                except:
                    radius = None
            else:
                radius = radius.split()[0].strip()
        elif "R_eff" in radius: # Some small bodies list an effective radius 
            radius = radius.split('R_eff')[1].strip(' =').split('+')[0].split()[0].strip()
        else: # Handles most major bodies
            radius = radius.split('=')[1].strip().split('+')[0].split()[0].strip() 
        try:
            radius = float(radius)
        except:
            radius = None
        return radius

    def get_rotrate(raw_response):
        if 'rot. rat' in raw_response.lower():  # or 'ROTPER' in raw_response.upper() or 'Orbital period' in raw_response:
            rotrate = [s for s in raw_response.split('\n') if 'rot. rat' in s.lower()]
            rotrate = rotrate[0].lower().split('rot. rat')[1].split('=')[1].strip().split('  ')[0].strip()
            try:
                rotrate = float(rotrate)
            except:
                rotrate = None
        elif 'ROTPER' in raw_response.upper():
            rotrate = [s for s in raw_response.split('\n') if 'ROTPER' in s.upper()] # Try the small body version
            rotrate = rotrate[0].split('ROTPER=')[1].strip()
            try:
                rotrate = 2*np.pi / (float(rotrate) * 3600)
            except:
                rotrate = None
        elif "Synchronous" in raw_response: # Satellites have this:
            rotrate = [s for s in raw_response.split('\n') if 'Orbital period' in s][0]
            rotrate = rotrate.split('Orbital period')[1].replace('~',' ').replace('d',' ').replace('=',' ').strip()
            rotrate = 2*np.pi / (float(rotrate) *  swiftest.JD2S) 
        elif "Orbital period" in raw_response:
            rotrate = [s for s in raw_response.split('\n') if 'Orbital period' in s][0]
            rotrate = rotrate.split('Orbital period')[1].split('=')[1].strip().split()[0].strip()
            rotrate = 2*np.pi / (float(rotrate) *  swiftest.JD2S)
        else:
            rotrate = None
        return rotrate

    def get_rotpole(jpl):
        RA = jpl.ephemerides()['NPole_RA'][0]
        DEC = jpl.ephemerides()['NPole_DEC'][0]
        
        if np.ma.is_masked(RA) or np.ma.is_masked(DEC):
            return np.array([0.0,0.0,1.0])

        rotpole = SkyCoord(ra=RA * u.degree, dec=DEC * u.degree,frame='icrs').transform_to('barycentricmeanecliptic').cartesian
         
        return np.array([rotpole.x.value, rotpole.y.value, rotpole.z.value])
  
    if ephemerides_start_date is None:
        ephemerides_start_date = constants.MINTON_BCL 
        
    if jpl is None:
        if not isinstance(id,list): 
            id = [id]
        jpl, altid, namelist = horizons_query(id=id[0],ephemerides_start_date=ephemerides_start_date,verbose=verbose,**kwargs)
    else: 
        if not isinstance(id,list):
            altid = [id]
        else:
            altid = id

    for i in altid:
        if jpl is None:
            jpl,_,namelist = horizons_query(id=i,ephemerides_start_date=ephemerides_start_date,verbose=verbose,**kwargs)
        namelist = [jpl.table['targetname'][0]]
        raw_response = jpl.vectors_async().text
        Rpl = get_radius(raw_response) 
        if Rpl is not None:
            Rpl *= 1e3
            rotpole = get_rotpole(jpl)
            break
        else:
            jpl = None
    Gmass = get_Gmass(raw_response)
    if Rpl is None or Gmass is None:
       mass = None
    else:
        Gmass *= 1e9  # JPL returns GM in units of km**3 / s**2, so convert to SI
        mass = Gmass / swiftest.GC
    if Rpl is not None:
        rotrate = get_rotrate(raw_response)
        if rotrate is None:
            rotrate = 0.0
        else:
            rotrate = np.rad2deg(rotrate)
    
        rot = rotpole*rotrate
    else:
        rot = np.full(3,np.nan)
    if Gmass is not None or Rpl is not None and verbose:
        if verbose:
            print(f"Physical properties found for {namelist[0]}") 
        
    return {'Gmass':Gmass,'mass':mass,'radius':Rpl,'rot':rot}


def horizons_query(id: str | int, 
                   ephemerides_start_date: str, 
                   exclude_spacecraft: bool=True, 
                   verbose: bool=False,
                   **kwargs: Any) -> Union[HorizonsClass | None, list | None, list | None]:
    """
    Queries JPL/Horizons for a body matching the id. If one is found, a HorizonsClass object is returned for the first object that
    matches the passed id string. If more than one match is found, a list of alternate ids is also returned. If no object is found
    then None is returned.

    Parameters
    ----------
    id : string
        A string identifying which body is requested from JPL/Horizons
    ephemerides_start_date : string
        Date to use when obtaining the ephemerides in the format YYYY-MM-DD.
    exclude_spacecraft: bool (optional) - Default True
        Indicate whether spacecraft ids should be exluded from the alternate id list
    verbose: bool (optional) - Default False
        Indicate whether to print messages about the query or not

    Returns
    -------
    jpl : HorizonsClass | None
        An astroquery.jplhorizons HorizonsClass object. Or None if no match was found.
    altid : string list | None
        A list of alternate ids if more than one object matches the list
    """
    
    def get_altid(errstr,exclude_spacecraft=True):
        """
        Parses the error message returned from a failed Horizons query. If the query failed because of an ambiguous id, then it will
        return a list of id values that could possibly match the query.
        not found

        Parameters
        ----------
        errstr : string
            The error message returned from the Horizons query
        exclude_spacecraft: bool (optional) - Default True
            Indicate whether spacecraft ids should be exluded from the alternate id list        

        Returns
        -------
        altid: string list | None
            A list of alternate ids if more than one object matches the list
        altname: string list | None
            A list of alternate names if more than one object matches the list
            
        """    
        if "ID" in errstr:
            altid = errstr.split('ID')[1]
            altid = altid.split('\n')[2:-1]
            altname = []
            for n,v in enumerate(altid):
                altid[n] = v.strip().split(' ')[0]
                altname.append(' '.join(v.strip().split(' ')[1:]).strip().split('  ')[0])
            if exclude_spacecraft:
                altname = [altname[i] for i,n in enumerate(altid) if int(n) > 0] 
                altid = [n for n in altid if int(n) > 0]
                
            return altid,altname 
        else:
            return None,None
        
    # Horizons date time internal variables
    tstart = datetime.date.fromisoformat(ephemerides_start_date)
    tstep = datetime.timedelta(days=1)
    tend = tstart + tstep
    ephemerides_end_date = tend.isoformat()
    ephemerides_step = '1d'
    
    try:
        jpl = Horizons(id=id, location='@sun',
                            epochs={'start': ephemerides_start_date, 'stop': ephemerides_end_date,
                                    'step': ephemerides_step},**kwargs)
        _=jpl.ephemerides()
        altid = [id]
        altname =[jpl.table['targetname'][0]]
    except Exception as e:
        altid,altname = get_altid(str(e))
        if altid is not None and len(altid) >0: # Return the first matching id
            id = altid[0]
            jpl = Horizons(id=id, location='@sun',
                        epochs={'start': ephemerides_start_date, 'stop': ephemerides_end_date,
                                'step': ephemerides_step})
            _=jpl.ephemerides()
        else:
            if verbose:
                warnings.warn(f"Could not find {id} in the JPL/Horizons system",stacklevel=2)
            return None,None,None
    if verbose:
        print(f"Found matching body: {altname[0]} ({altid[0]})") 
        if len(altid) > 1:
            print("    Alternate matching bodies:")
            for i,v in enumerate(altid):
                if i > 0:
                    print(f"    {altname[i]} ({v})")
        
    return jpl,altid,altname
    
    
def get_solar_system_body(name: str,
                          ephemeris_id: str | None = None,
                          ephemerides_start_date : str = constants.MINTON_BCL,
                          central_body_name: str = "Sun",
                          verbose: bool = True,
                          **kwargs: Any) -> dict | None:
    """
    Initializes a Swiftest dataset containing the major planets of the Solar System at a particular data from JPL/Horizons

    Parameters
    ----------
    name  : string
        Name of body to add to Dataset. If `id` is not supplied, this is also what will be searched for in the JPL/Horizon's database.
        The first matching body is found (for major planets, this is the barycenter of a planet-satellite system)
    ephemeris_id : string (optional)
        If passed, this is passed to Horizons instead of `name`. This can be used to find a more precise body than given by `name`.
    ephemerides_start_date : string
        Date to use when obtaining the ephemerides in the format YYYY-MM-DD. Default is constants.MINTON_BCL
    central_body_name : string
        Name of the central body to use when calculating the relative position and velocity vectors. Default is "Sun"
    verbose : bool
        Indicates whether to print messages about the query or not. Default is True
    **kwargs: Any
        Additional keyword arguments to pass to the query method (see https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html)

    Returns
    -------
    A dictionary containing the following elements
    
    name : string
        Name of the body
    rh : (3,) array of np.float64
        Position vector array relative to the central body in m.
    vh : (3,) array of np.float64
        Velocity vector array relative to the central body in m/s.
    Gmass : np.float64
        G*mass values if these are massive bodies in m^3/s^2
    mass : np.float64
        Mass values if these are massive bodies in kg
    radius : np.float64
        Radius values if these are massive bodies in m
    rhill : np.float64 
        The Hill's radius values if these are massive bodies in m
    Ip : (3,) array of np.float64
        Principal axes moments of inertia vectors if these are massive bodies.
    rot : (3,) array of np.float
        Rotation rate vectors if these are massive bodies in deg/s
    j2rp2 : np.float64
        J_2R^2 value for the body if known
    j4rp4 : np.float64
        J_4R^4 value for the body if known
         
    Notes
    --------
    When passing `name` == "Earth" or `name` == "Pluto", it a body is generated that has initial conditions matching the system
    barycenter and mass equal to the sum of Earth+Moon or Pluto+Charon. To obtain initial conditions for either Earth or Pluto alone,
    pass `ephemeris_id` == "399" for Earth or `id` == "999" for Pluto. 
    """
    
    planetIpz = { # Only the polar moments of inertia are used currently. Where the quantity is unkown, we just use the value of a sphere = 0.4
        'Sun' : np.longdouble(0.070),
        'Mercury' : np.longdouble(0.346),
        'Venus': np.longdouble(0.4),
        'Earth': np.longdouble(0.3307),
        'Mars': np.longdouble(0.3644),
        'Jupiter': np.longdouble(0.2756),
        'Saturn': np.longdouble(0.22),
        'Uranus': np.longdouble(0.23),
        'Neptune': np.longdouble(0.23),
        'Pluto': np.longdouble(0.4)
        }
   
    # J2 and J4 for the major planets are from From Murray & Dermott (1999) Table A.4. 
    # The Sun is from Mecheri et al. (2004), using Corbard (b) 2002 values (Table II)
    planetJ2 = {
        'Sun' : np.longdouble(2.198e-7),
        'Mercury' : np.longdouble(60.0 * 1e-6),
        'Venus' : np.longdouble(4.0 * 1e-6),
        'Earth' : np.longdouble(1083.0 * 1e-6),
        'Mars' : np.longdouble(1960.0 * 1e-6),
        'Jupiter': np.longdouble(14736.0 * 1e-6),
        'Saturn': np.longdouble(16298.0 * 1e-6),
        'Uranus': np.longdouble(3343.0 * 1e-6),
        'Neptune': np.longdouble(3411.0 * 1e-6),
    }
    planetJ4 = { 
        'Sun' : np.longdouble(-4.805e-9),
        'Mercury' : np.longdouble(0.0),
        'Venus' : np.longdouble(2.0 * 1e-6),
        'Earth' : np.longdouble(-2.0 * 1e-6),
        'Mars' : np.longdouble(-19.0 * 1e-6),
        'Jupiter': np.longdouble(-587.0 * 1e-6),
        'Saturn': np.longdouble(-915.0 * 1e-6),
        'Uranus': np.longdouble(-29.0 * 1e-6),
        'Neptune': np.longdouble(-35.0 * 1e-6),
    }

    # Unit conversion factors
    DCONV = swiftest.AU2M 
    VCONV = (swiftest.AU2M / swiftest.JD2S) 
    
    rh = np.full(3,np.nan)
    vh = np.full(3,np.nan)
    Ip = np.full(3,np.nan)
    rot = np.full(3,np.nan)
    rhill = None
    Gmass = None
    mass = None
    Rpl = None
    j2rp2 = None
    j4rp4 = None
   
    if name == "Sun" or ephemeris_id == "0": # Create central body
        if verbose:
            print("Creating the Sun as a central body")
        # Central body value vectors
        rotpoleSun = SkyCoord(ra=286.13 * u.degree, dec=63.87 * u.degree).cartesian
        rot = (360.0 / 25.05) / constants.JD2S  * rotpoleSun           
        rot = np.array([rot.x.value, rot.y.value, rot.z.value])
        Gmass = swiftest.GMSun
        mass = Gmass / swiftest.GC
        Rpl = swiftest.RSun 
        rh = np.array([0.0, 0.0, 0.0])
        vh = np.array([0.0, 0.0, 0.0])
    else: # Fetch solar system ephemerides from Horizons
        if ephemeris_id is None:
            ephemeris_id = name
           
        if verbose: 
            print(f"Fetching ephemerides data for {ephemeris_id} from JPL/Horizons")
        
        jpl,altid,altname = horizons_query(ephemeris_id,ephemerides_start_date,verbose=verbose,**kwargs)
        if jpl is not None:
            if verbose:
                print(f"Found ephemerides data for {altname[0]} ({altid[0]}) from JPL/Horizons")
            if name == None:
                name = altname[0]
        else:
            return None
        
        if central_body_name != "Sun":
            jplcb, altidcb, _ = horizons_query(central_body_name,ephemerides_start_date,verbose=verbose,**kwargs)
            GMcb = get_solar_system_body_mass_rotation(altidcb,jplcb)['Gmass']
            cbrx = jplcb.vectors()['x'][0] * DCONV
            cbry = jplcb.vectors()['y'][0] * DCONV
            cbrz = jplcb.vectors()['z'][0] * DCONV
            cbvx = jplcb.vectors()['vx'][0] * VCONV
            cbvy = jplcb.vectors()['vy'][0] * VCONV
            cbvz = jplcb.vectors()['vz'][0] * VCONV
            cbrh = np.array([cbrx,cbry,cbrz])
            cbvh = np.array([cbvx,cbvy,cbvz])
        else:
            GMcb = swiftest.GMSun 
            cbrh = np.zeros(3)
            cbvh = np.zeros(3)
    
        rx = jpl.vectors()['x'][0] * DCONV
        ry = jpl.vectors()['y'][0] * DCONV
        rz = jpl.vectors()['z'][0] * DCONV
        vx = jpl.vectors()['vx'][0] * VCONV
        vy = jpl.vectors()['vy'][0] * VCONV
        vz = jpl.vectors()['vz'][0] * VCONV

        rh = np.array([rx,ry,rz]) - cbrh
        vh = np.array([vx,vy,vz]) - cbvh

        Gmass,_,Rpl,rot = get_solar_system_body_mass_rotation(altid,jpl,verbose=verbose,**kwargs).values()
        # If the user inputs "Earth" or Pluto, then the Earth-Moon or Pluto-Charon barycenter and combined mass is used. 
        # To use the Earth or Pluto alone, simply pass "399" or "999", respectively to name
        if name == "Earth":
            if verbose:
                print("Combining mass of Earth and the Moon")
            Gmass_moon = get_solar_system_body_mass_rotation(["301"],verbose=verbose,**kwargs)['Gmass']
            Gmass += Gmass_moon
        elif name == "Pluto":
            if verbose:
                print("Combining mass of Pluto and Charon")
            Gmass_charon = get_solar_system_body_mass_rotation(["901"],verbose=verbose,**kwargs)['Gmass']
            Gmass += Gmass_charon 
        
        if Gmass is not None:
            rhill = jpl.elements()['a'][0] * DCONV * (Gmass / (3*GMcb))**(1.0/3.0)
            mass = Gmass / swiftest.GC
            
    if name in planetIpz:
        Ip = np.array([0.0, 0.0, planetIpz[name]])
    else:
        Ip = np.array([0.4, 0.4, 0.4])
               
    if name in planetJ2:
        j2rp2 = planetJ2[name] * Rpl**2 
        j4rp4 = planetJ4[name] * Rpl**4
        
    return {'name':name,
            'rh':rh,
            'vh':vh,
            'Gmass':Gmass,
            'mass' : mass,
            'radius': Rpl,
            'rhill': rhill,
            'Ip': Ip,
            'rot': rot,
            'j2rp2':j2rp2,
            'j4rp4':j4rp4}


#!/usr/bin/env python3
"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""
import swiftest
import unittest
import os
import tempfile
import astropy.constants as const
import numpy as np

class TestSwiftestUnits(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.tmpdir=tempfile.TemporaryDirectory()
        self.simdir = self.tmpdir.name
        
    def tearDown(self):
        # Clean up temporary directory
        self.tmpdir.cleanup() 
        
    def test_unit_conversions(self):
        """
        Test setting standard and custom units and checks that the conversions are correct
        """
        sim = swiftest.Simulation(simdir=self.simdir)
        MSun = const.M_sun.value
        sim.set_parameter(MU="Msun")
        self.assertEqual(sim.MU_name, "MSun")
        self.assertAlmostEqual(sim.MU2KG/MSun, np.longdouble(1.0))
        self.assertAlmostEqual(sim.KG2MU*MSun, np.longdouble(1.0))
        self.assertEqual(sim.MU2KG,sim.param['MU2KG']) 

        sim.set_parameter(MU2KG=1e-3)
        self.assertEqual(sim.MU_name, "MU")
        self.assertEqual(sim.MU2KG, np.longdouble(1e-3))
        self.assertEqual(sim.KG2MU, np.longdouble(1000.0))
        self.assertEqual(sim.MU2KG,sim.param['MU2KG']) 
       
        sim.set_parameter(DU="cm")
        self.assertEqual(sim.DU_name, "cm")
        self.assertEqual(sim.DU2M, np.longdouble(1e-2))
        self.assertEqual(sim.M2DU, np.longdouble(100.0))
        
        sim.set_parameter(DU="km")
        self.assertEqual(sim.DU_name, "km")
        self.assertEqual(sim.DU2M, np.longdouble(1e3))
        
         
if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()


from __future__ import annotations
import xarray as xr
import numpy as np
from scipy.spatial.transform import Rotation as R
from .constants import *

class SwiftestDataArray(xr.DataArray):
    """
    N-dimensional ``xarray.DataArray``-like array. Inherits from ``xarray.DataArray`` and has its own set of methods and attributes 
    specific to the Swiftest project

    Parameters
    ----------------
    *args:
        Arguments for the ``xarray.DataArray`` class
    **kwargs:
        Keyword arguments for the ``xarray.DataArray`` class

    Notes
    -----
    See `xarray.DataArray <https://docs.xarray.dev/en/stable/generated/xarray.DataArray.html>`__
    for further information about DataArrays.
    """
    
    __slots__ = ()

    def __init__(self, *args,**kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _construct_direct(cls, *args, **kwargs):
        """Override to make the result a ``swiftest.SwiftestDataArray`` class."""
        return cls(xr.DataArray._construct_direct(*args, **kwargs))

    def to_dataset(self): 
        """Converts a ``SwiftestDataArray`` into a ``SwiftestDataset`` with a single data
        variable."""
        xrds = super().to_dataset()
        return SwiftestDataset(xrds)
    
    def magnitude(self, name: str | None = None): 
        """
        Computes the magnitude of a vector quantity. Note: The DataArray must have the "space" dimension. 
        
        Parameters
        ----------
        name : str, optional
            Name of the new DataArray. By default, the string "_mag" is appended to the original name.
        
        Returns
        -------
        mag : SwiftestDataArray
            DataArray containing the magnitude of the vector quantity 
        """
        dim = "space"
        ord = None
        
        if dim not in self.dims:
            raise ValueError(f"Dimension {dim} not found in DataArray")
        if name is None and isinstance(self.name, str):
                name = self.name + "_mag"
        da = xr.apply_ufunc(
            np.linalg.norm, self.where(~np.isnan(self)), input_core_dims=[[dim]], kwargs={"ord": ord, "axis": -1}, dask="allowed"
        )
        
        da = da.rename(name)
        return SwiftestDataArray(da)


    def rotate(self, rotation):
        """
        Rotates a vector quantity using a rotation matrix. The DataArray must have the "space" dimension. 
        
        Parameters
        ----------
        rotation : (3) float array
            Rotation vector
        """
       
        if "space" not in self.dims:
            raise ValueError("DataArray must have a 'space' dimension")
        
        # Define a function to apply the rotation, which will be used with apply_ufunc
        def apply_rotation(vector, rotation):
            if not rotation.single: 
                # If 'rotation' is a stack of rotations, apply each rotation sequentially
                for single_rotation in rotation:
                    vector = single_rotation.apply(vector)
                return vector
            else:
                # If 'rotation' represents a single rotation, apply it directly
                return rotation.apply(vector)
        
        da = xr.apply_ufunc(
            apply_rotation,
            self,
            kwargs={'rotation': rotation},
            input_core_dims=[['space']],
            output_core_dims=[['space']],
            vectorize=True,
            dask='parallelized',
            output_dtypes=[self.dtype]
        )
        return SwiftestDataArray(da) 
    

class SwiftestDataset(xr.Dataset):
    """
    A ``xarray.Dataset``-like, multi-dimensional, in memory, array database.  Inherits from ``xarray.Dataset`` and has its own set of
    methods and attributes specific to the Swiftest project.

    Parameters
    ----------------
    *args:
        Arguments for the ``xarray.Dataset`` class
    **kwargs:
        Keyword arguments for the ``xarray.Dataset`` class

    Notes
    -----
    See `xarray.Dataset <https://docs.xarray.dev/en/stable/generated/xarray.Dataset.html>`__
    for further information about Datasets.
    """
    __slots__ = ()
    def __init__(
                 self,
                 *args,
                 **kwargs,
                ):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        """Override to make sure the result is an instance of
        ``swiftest.SwiftestDataArray`` or ``swiftest.SwiftestDataset``."""

        value = super().__getitem__(key)

        if isinstance(value, xr.DataArray):
            value = SwiftestDataArray(value)
        elif isinstance(value, xr.Dataset):
            value = SwiftestDataset(value)

        return value

    def _calculate_binary_op(self, *args, **kwargs):
        """Override to make the result a complete instance of ``swiftest.SwiftestDataset``."""
        ds = super()._calculate_binary_op(*args, **kwargs)
        if not isinstance(ds, SwiftestDataset):
            ds = SwiftestDataset(ds)
        ds = SwiftestDataset(ds)

        return ds

    def _construct_dataarray(self, name) -> SwiftestDataArray:
        """Override to make the result an instance of ``swiftest.SwiftestDataArray``."""
        xarr = super()._construct_dataarray(name)
        return SwiftestDataArray(xarr)

    @classmethod
    def _construct_direct(cls, *args, **kwargs):
        """Override to make the result an ``swiftest.SwiftestDataset`` class."""

        return cls(xr.Dataset._construct_direct(*args, **kwargs))


    def _replace(self, *args, **kwargs):
        """Override to make the result a complete instance of ``swiftest.SwiftestDataset``."""
        ds = super()._replace(*args, **kwargs)

        if not isinstance(ds, SwiftestDataset):
            ds = SwiftestDataset(ds)

        return ds

    @classmethod
    def from_dataframe(cls, dataframe):
        """Override to make the result a ``swiftest.SwiftestDataset`` class."""

        return cls(
            {col: ("index", dataframe[col].values) for col in dataframe.columns},
            coords={"index": dataframe.index},
        )

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Override to make the result a ``swiftest.SwiftestDataset`` class."""

        return cls(
            {key: ("index", val) for key, val in data.items()},
            coords={"index": range(len(next(iter(data.values()))))},
            **kwargs,
        )


    def rotate(self, rotvec=None, pole=None, skip_vars=['space','Ip']):
        """
        Rotates the coordinate system such that the z-axis is aligned with an input pole.  The new pole is defined by the input vector. 
        This will change all variables in the Dataset that have the "space" dimension, except for those passed to the skip_vars parameter.
        
        Parameters
        ----------
        ds : SwiftestDataset
            Dataset containing the vector quantity
        rotvec: (N,3) or (3,) float array
            Rotation vector
        pole : (3) float array
            New pole vector
        skip_vars : list of str, optional
            List of variable names to skip. The default is ['space','Ip'].
            
        Returns
        -------
        ds : SwiftestDataset
            Dataset with the new pole vector applied to all variables with the "space" dimension
            
        Notes
        -----
        You can pass either rotvec or pole, but not both. If both, or none, are passed, the function will raise an exception. 
        """
        if rotvec is not None and pole is not None:
            raise ValueError("You can only pass either rotvec or pole, but not both")
        if rotvec is None and pole is None:
            raise ValueError("You must pass either rotvec or pole")
         
        if 'space' not in self.dims:
            raise ValueError("Dataset must have a 'space' dimension")    
        
        # Verify that the new pole is a 3-element array
        if pole is not None:
            if len(pole) != 3:
                raise ValueError("Pole vector must be a 3-element array")
    
            # Normalize the new pole vector to ensure it is a unit vector
            pole_mag = np.linalg.norm(pole)
            unit_pole = pole / pole_mag
        
            # Define the original and target vectors
            target_vector = np.array([0, 0, 1])  # Rotate so that the z-axis is aligned with the new pole
            original_vector = unit_pole.reshape(1, 3)  
        
            # Use align_vectors to get the rotation that aligns the z-axis with Mars_rot
            rotvec, _ = R.align_vectors(target_vector, original_vector)
        elif rotvec is not None:
            rotvec = np.asarray(rotvec)
            if (rotvec.shape[-1]) != 3:
                raise ValueError("Rotation vector must be a 3-element array")
            
            rotvec = R.from_rotvec(rotvec) 

        # Loop through each variable in the dataset and apply the rotation if 'space' dimension is present
        for var in self.variables:
            if 'space' in self[var].dims and var not in skip_vars:
                self[var] = self[var].rotate(rotvec)
                
        return self
    
        
    def el2xv(self, GMcb: xr.DataArray | float | None = None) -> SwiftestDataset:
        """
        Converts a Dataset's orbital elements to Cartesian state vectors. The DataArray must have the appropriate dimensions for orbital elements.
        
        Parameters
        ----------
        GMcb : xr.DataArray or float
            Gravitational parameter of the central body
        
        Returns
        -------
        SwiftestDataset
            Dataset containing the computed state vectors (position 'rh' and velocity 'vh').
        """
        from .core import el2xv  # Assuming el2xv is implemented in the .core module

        if 'space' not in self.dims:
            raise ValueError("Dataset must have a 'space' dimension")
        
        required_vars = ['a', 'e', 'inc', 'capom', 'omega', 'capm']
        for var in required_vars:
            if var not in self.variables:
                raise ValueError(f"Dataset must have '{var}' variables")
        
        # Identify the index dimension
        if 'id' in self.dims:
            index_dim = 'id'
        elif 'name' in self.dims:
            index_dim = 'name'
        else:
            raise ValueError("Dataset must have an 'id' or 'name' dimension")
        
        if GMcb is None:
            if 'Gmass' not in self:
                raise ValueError("Dataset must have a 'Gmass' variable for the central body")
            if 'particle_type' in self.variables:
                GMcb = self['Gmass'].where(self['particle_type'] == CB_TYPE_NAME, drop=True)
            else:
                GMcb = self['Gmass'].where(self['id'] == 0, drop=True)
            if GMcb.size != 1:
                raise ValueError("Dataset must have a single central body") 
        
        if isinstance(GMcb, xr.DataArray):
            if 'id' in GMcb.dims:
                GMcb = GMcb.isel(id=0)
            elif 'name' in GMcb.dims:
                GMcb = GMcb.isel(name=0)
                
        if isinstance(GMcb, xr.DataArray):
            if 'id' in GMcb.dims:
                GMcb = GMcb.isel(id=0)
            elif 'name' in GMcb.dims:
                GMcb = GMcb.isel(name=0)
        else:
            GMcb = xr.DataArray(data = GMcb)
            
        for dim in self.a.dims:
            if dim not in GMcb.dims:
                GMcb = GMcb.expand_dims(dim={dim: self[dim]}) 
                
        if 'Gmass' in self:
            mu = xr.where(self['Gmass'] > 0.0, GMcb + self['Gmass'], GMcb)
        else:
            mu = GMcb
        
        # Prepare the orbital elements for the function call
        a = self['a'].astype(np.float64)
        e = self['e'].astype(np.float64)
        inc = self['inc'].astype(np.float64)
        capom = self['capom'].astype(np.float64)
        omega = self['omega'].astype(np.float64)
        capm = self['capm'].astype(np.float64)
        mu = mu.astype(np.float64)

        # Use apply_ufunc to convert orbital elements back to state vectors
        rh, vh = xr.apply_ufunc(
            el2xv,  # Function to apply
            mu, a, e, inc, capom, omega, capm,  # Inputs
            input_core_dims=[[index_dim], [index_dim], [index_dim], [index_dim], [index_dim], [index_dim], [index_dim]],  # Core dimensions for each input
            output_core_dims=[[index_dim, 'space'], [index_dim, 'space']],  # Core dimensions for outputs (position and velocity vectors)
            vectorize=True,  # Automatically vectorize over non-core dimensions
            dask="parallelized",  # Enable parallelized computation for Dask arrays, if applicable
            output_dtypes=[np.float64, np.float64]  # Expected data types for outputs
        )

        # Create a new Dataset with the state vectors
        new_vars = {'rh': rh, 'vh': vh}
        dataset = xr.Dataset(new_vars)
        if "name" in dataset.variables:
            dataset = dataset.drop_vars("name")        
        dsnew = xr.merge([dataset, self], compat="override")
        
        return SwiftestDataset(dsnew)
     
                
    def xv2el(self, GMcb: xr.DataArray | float | None = None) -> SwiftestDataset:
        """
        Converts A Dataset's Cartesian state vectors to orbital elements. The DataArray must have the "space" dimension. 
        
        Parameters
        ----------
        GMcb : xr.DataArray or float 
            Gravitational parameter of the central body
        
        Returns
        -------
        SwiftestDataset
            Dataset containing the computed orbital elements.
        """
        from .core import xv2el
        if 'space' not in self.dims:
            raise ValueError("Dataset must have a 'space' dimension")
        if 'rh' not in self.variables or 'vh' not in self.variables:
            raise ValueError("Dataset must have 'rh' and 'vh' variables")
       
        if 'id' in self.dims:
            index_dim = 'id'
        elif 'name' in self.dims:
            index_dim = 'name'
        else:
            raise ValueError("Dataset must have an 'id' or 'name' dimension")
        
        if GMcb is None:
            if 'Gmass' not in self:
                raise ValueError("Dataset must have a 'Gmass' variable for the central body")
            if 'particle_type' in self.variables:
                GMcb = self['Gmass'].where(self['particle_type'] == CB_TYPE_NAME, drop=True)
            else:
                GMcb = self['Gmass'].where(self['id'] == 0, drop=True)
            if GMcb.isel(time=0).size != 1:
                raise ValueError("Dataset must have a single central body") 
        
        if isinstance(GMcb, xr.DataArray):
            if 'id' in GMcb.dims:
                GMcb = GMcb.isel(id=0)
            elif 'name' in GMcb.dims:
                GMcb = GMcb.isel(name=0)
        else:
            GMcb = xr.DataArray(data = GMcb)
            
        for dim in self.rh.dims:
            if dim not in GMcb.dims and dim not in ['space']:
                GMcb = GMcb.expand_dims(dim={dim: self[dim]}) 
                
        if 'Gmass' in self:
            mu = xr.where(self['Gmass'] > 0.0, GMcb + self['Gmass'], GMcb)
        else:
            mu = GMcb
            
        # Prepare the cartesian state vectorsfor the function call
        rh = self['rh'].astype(np.float64)
        vh = self['vh'].astype(np.float64)  
        mu = mu.astype(np.float64)
        
        result = xr.apply_ufunc(
            xv2el,  
            mu, rh, vh,
            input_core_dims=[[index_dim], [index_dim, 'space'], [index_dim, 'space']],  
            output_core_dims=[[index_dim]] * 11,  
            vectorize=True,  
            dask="parallelized",  
            output_dtypes=[np.float64] * 11  
        )
        
        varnames = ['a', 'e', 'inc', 'capom', 'omega', 'capm', 'varpi', 'lam', 'f', 'cape', 'capf']
        dataset = xr.Dataset({var: result[i] for i, var in enumerate(varnames)})
        if "name" in dataset.variables:
            dataset = dataset.drop_vars("name")        
        dsnew = xr.merge([dataset, self], compat="override")
        
        return SwiftestDataset(dsnew)




#!/usr/bin/env python3
"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""
import swiftest
import unittest
import os
import tempfile
import numpy as np
class TestCollisions(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.tmpdir=tempfile.TemporaryDirectory()
        self.simdir = self.tmpdir.name
        
    def tearDown(self):
        # Clean up temporary directory
        self.tmpdir.cleanup() 
        
    def test_solar_impact(self):
        '''
        Tests that impacts into the central body work correctly
        '''
        sim = swiftest.Simulation(simdir=self.simdir,compute_conservation_values=True, rotation=True, collision_model='merge')

        # Add the modern planets and the Sun using the JPL Horizons Database.
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune","Pluto"])
        runargs = {"tstart":0.0, "tstop":5e-2, "dt":0.0001, "istep_out":1, "dump_cadence":0}

        density  = 3000.0 * sim.KG2MU / sim.M2DU**3

        # Make a massive body with a periapsis inside the Sun's radius
        q = 0.01 * swiftest.RSun * sim.M2DU
        a = 0.1
        e = 1.0 - q / a
        M = 0.1 * sim.init_cond.sel(name="Earth")['mass'].values[0]
        R = (3 * M  / (4 * np.pi * density)) ** (1.0 / 3.0)
        rot = 4 * sim.init_cond.sel(name="Earth")['rot']
        sim.add_body(name="Sundiver", a=a, e=e, inc=0.0, capom=0.0, omega=0.0, capm=180.0, mass=M, radius=R, Ip=[0.4,0.4,0.4], rot=rot)
        
        for mtiny in [2*M,0.5*M]: 
            sim.clean()
            sim.run(**runargs,mtiny=mtiny,integrator="symba")
        
            # Check that the collision actually happened
            self.assertEqual(sim.collisions.collision_id.size,1) 
        
            # Check that angular momentum is conserved
            ds=sim.collisions.sel(collision_id=1)
            ds['Ltot']=ds.L_orbit+ds.L_rot
            ds['Ltot_mag']=ds.Ltot.magnitude()
            dLtot=ds.Ltot_mag.diff('stage').values[0]
            self.assertLess(dLtot,1e-6, msg=f"Angular momentum not conserved: {dLtot}")
        
            # Check that energy was lost
            dEtot=ds.TE.diff('stage').values[0]
            self.assertLess(dEtot,0, msg=f"Energy not lost: {dEtot}")
        
        # Test that massive bodies can be discarded in RMVS
        sim.run(**runargs,integrator="rmvs")
        # Check that the collision actually happened
        self.assertEqual(sim.collisions.collision_id.size,1, msg="Collision not detected in RMVS") 
       
        # Test that test particles are discarded in RMVS and SyMBA 
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune","Pluto"])
        sim.add_body(name="Sundiver", a=a, e=e, inc=0.0, capom=0.0, omega=0.0, capm=180.0)
        for integrator in ['rmvs','symba','whm','helio']:
            sim.clean()
            sim.run(**runargs,integrator=integrator)
            self.assertEqual(sim.collisions.collision_id.size,1, msg=f"Collision not detected in {integrator}") 
            self.assertEqual(sim.collisions.sel(collision_id=1).regime.values, 'Central Body Impact', msg=f"{integrator}: Wrong regime: {sim.collisions.sel(collision_id=1).regime.values}") 
        
        return 
    
    def test_escape(self):
        '''
        Tests that escaping bodies are handled correctly
        '''
        sim = swiftest.Simulation(simdir=self.simdir,compute_conservation_values=True, rotation=True)

        # Add the modern planets and the Sun using the JPL Horizons Database.
        sim.add_solar_system_body(["Sun","Jupiter","Saturn","Uranus","Neptune","Pluto"])

        density  = 3000.0 * sim.KG2MU / sim.M2DU**3

        # Make a hyperbolic body
        a = 0.1
        e = 1.1
        M = 1e-4 * swiftest.MEarth * sim.KG2MU
        R = (3 * M  / (4 * np.pi * density)) ** (1.0 / 3.0)
        rot = 2 * sim.init_cond.sel(name="Saturn")['rot']
        sim.add_body(name="Escapee", a=a, e=e, inc=0.0, capom=0.0, omega=45.0, capm=180.0, mass=M, radius=R, Ip=[0.4,0.4,0.4], rot=rot)
        
        runargs = {"tstart":0.0, "tstop":1000.0, "dt":0.05, "istep_out":200, "dump_cadence":0}
       
        for mtiny in [2*M,0.5*M]: 
            # Try with semi-interacting body
            sim.clean()
            sim.run(**runargs, mtiny=mtiny, integrator="symba")
            
            # Check that the escape event was recorded
            self.assertEqual(sim.collisions.collision_id.size,1) 
            self.assertEqual(sim.collisions.sel(collision_id=1).regime.values, 'Ejected', msg=f"mtiny/M: {mtiny/M}: Wrong regime: {sim.collisions.sel(collision_id=1).regime.values}") 
       
            # Check that angular momentum is conserved
            ds=sim.collisions.sel(collision_id=1)
            ds['Ltot']=ds.L_orbit+ds.L_rot
            ds['Ltot_mag']=ds.Ltot.magnitude()
            dLtot=ds.Ltot_mag.diff('stage').values[0] 
            self.assertLess(dLtot,1e-6, msg=f"Mtiny/M: {mtiny/M} Angular momentum not conserved: {dLtot}")
            
            # Check that energy was lost
            dEtot=ds.TE.diff('stage').values[0]
            self.assertLessEqual(dEtot,0, msg=f"Energy not lost: {dEtot}")
        
        # Test that massive bodies can be discarded in all integrators
        for integrator in ['rmvs','symba','whm','helio']:
            sim.clean()
            sim.run(**runargs, integrator=integrator)
            # Check that the collision actually happened
            self.assertEqual(sim.collisions.collision_id.size,1, msg=f"Collision not detected in {integrator}") 
            
        # Test that test particles are discarded in all integrators
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(["Sun","Jupiter","Saturn","Uranus","Neptune","Pluto"])
        sim.add_body(name="Escapee", a=a, e=e, inc=0.0, capom=0.0, omega=45.0, capm=180.0)
        # Test that massive bodies can be discarded in all integrators
        for integrator in ['rmvs','symba','whm','helio']:
            sim.clean()
            sim.run(**runargs, integrator=integrator)
            # Check that the collision actually happened
            self.assertEqual(sim.collisions.collision_id.size,1, msg=f"Collision not detected in {integrator}")         
            self.assertEqual(sim.collisions.sel(collision_id=1).regime.values, 'Ejected', msg=f"{integrator}: Wrong regime: {sim.collisions.sel(collision_id=1).regime.values}") 
        
        return 
        
    def test_merge(self):
        """
        Tests that merging bodies are handled correctly
        """
        sim = swiftest.Simulation(simdir=self.simdir,compute_conservation_values=True, rotation=True, collision_model='merge') 
        sim.add_solar_system_body("Sun")
        
        name = ["Target","Projectile"]
        rh = [np.array([1.0, -5.0e-05, 0.0]),
               np.array([1.0,  5.0e-05 ,0.0])]
        vh = [np.array([ 0.00,  6.280005, 0.0]),
               np.array([ 0.00,  3.90,     0.0])]
        rot = [np.array([0.0, 0.0, 1.0e5]),
              np.array([0.0, 0.0, -5e5])]
        Gmass = [1e-7, 1e-9]
        density  = 3000.0 * sim.KG2MU / sim.M2DU**3
        radius = [((GM/sim.GU)/(4./3.*np.pi*density))**(1./3.) for GM in Gmass]
        runargs = {"tstart":0.0, "tstop":2e-3, "dt":5e-4, "istep_out":1, "dump_cadence":0}
        
        # Test that massive body merge works in SyMBA
        sim.add_body(name=name, rh=rh, vh=vh, rot=rot, Gmass=Gmass, radius=radius)   
        
        for gmtiny in [2*Gmass[1],0.5*Gmass[1]]: 
            sim.clean()
            sim.run(**runargs,gmtiny=gmtiny,integrator="symba")
        
            # Check that the collision actually happened
            self.assertEqual(sim.collisions.collision_id.size,1) 
       
            # Check that the collision was a merge     
            self.assertEqual(sim.collisions.sel(collision_id=1).regime.values, 'Merge', msg=f"Wrong regime: {sim.collisions.sel(collision_id=1).regime.values}") 
        
            # Check that angular momentum is conserved
            ds=sim.collisions.sel(collision_id=1)
            ds['Ltot']=ds.L_orbit+ds.L_rot
            ds['Ltot_mag']=ds.Ltot.magnitude()
            dLtot=ds.Ltot_mag.diff('stage').values[0]
            self.assertAlmostEqual(dLtot,0,places=8, msg=f"Angular momentum not conserved: {dLtot}")
        
            # Check that energy was lost
            dEtot=ds.TE.diff('stage').values[0]
            self.assertLess(dEtot,0, msg=f"Energy not lost: {dEtot}")        
        
    def test_multi_collision(self):
        """
        Tests that multiple collisions are handled correctly
        """              
        
        sim = swiftest.Simulation(simdir=self.simdir, collision_model='merge')
        

        sim.add_solar_system_body(['Sun', 'Earth'])

        # add bodies that both impact the CB at the same time
        sim.add_body(name = ['close'], rh = [0, 0, sim.param['CHK_RMIN'] * 1.1], vh = [0, 0, 0.0001])
        sim.add_body(name = ['close2'], rh = [0, 0, sim.param['CHK_RMIN'] * 1.11], vh = [0, 0, 0.0001])

        try:
            sim.run(tstop = 10, dt = 0.01, tstep_out = 1)
        except Exception as e:
            self.fail(f"Multiple collisions not handled correctly: {e}")
            
        self.assertEqual(sim.collisions.collision_id.size,2)
        
    def test_multi_pltp_collision(self):
        """
        Tests that multiple pl-tp collisions are handled correctly
        """
        
        sim = swiftest.Simulation(simdir=self.simdir) #

        sim.add_solar_system_body(['Sun', 'Earth'])
        ncoll = 10
        nnotcoll = 10

        # Create bodies that will collide
        rh_E = sim.init_cond.sel(time=0, name='Earth')['rh'].values
        vh_E = sim.init_cond.sel(time=0, name='Earth')['vh'].values
        rad_E = sim.init_cond.sel(time=0, name='Earth')['radius'].values.item()

        rh_tp = []
        vh_tp = []
        for i in range(ncoll): 
            rh_tp.append(rh_E + [0,-20*(i+1)*rad_E, 0])
            vh_tp.append(vh_E + [0.0, 1e-2, 0])

        sim.add_body(rh=rh_tp, vh=vh_tp)

        # Create bodies that will not collide
        a_tp = []
        a_tp = [10.0+0.1*i for i in range(nnotcoll)]
        sim.add_body(a=a_tp)
        
        try:
            sim.run(tstop = 100, dt = 0.001, tstep_out = 10, integrator='rmvs')
        except Exception as e:
            self.fail(f"Multiple pl-tp collisions not handled correctly: {e}")

        self.assertEqual(sim.collisions.collision_id.size,ncoll)


if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()


#!/usr/bin/env python3
"""
Copyright 2024 - The Minton Group at Purdue University
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest.
If not, see: https://www.gnu.org/licenses.
"""
import swiftest
import unittest
import os
import tempfile
import numpy as np

# Colliion initial conditions taken from Fragmentation_Movie.py in the example directory

collision_type = ["disruption_headon", "disruption_off_axis", "supercatastrophic_headon", "supercatastrophic_off_axis","hitandrun_disrupt", "hitandrun_pure", "merge", "merge_spinner"]

# ----------------------------------------------------------------------------------------------------------------------
# To increase the number of bodies generated in each collision type, decrease the value of the corresponding nfrag_reduction number
# ----------------------------------------------------------------------------------------------------------------------
nfrag_reduction = {"disruption_headon" : 10.0,
        "disruption_off_axis"         : 10.0,
        "supercatastrophic_headon"    : 10.0,
        "supercatastrophic_off_axis"  : 10.0,
        "hitandrun_disrupt"           : 10.0,
        "hitandrun_pure"              : 1.0,
        "merge"                       : 1.0,
        "merge_spinner"               : 1.0,
        }


# These initial conditions were generated by trial and error
names = ["Target","Projectile"]
pos_vectors = {"disruption_headon"         : [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05 ,0.0])],
            "disruption_off_axis"        : [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05 ,0.0])],
            "supercatastrophic_headon":   [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05, 0.0])],
            "supercatastrophic_off_axis": [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05, 0.0])],
            "hitandrun_disrupt"         : [np.array([1.0, -4.2e-05, 0.0]),
                                            np.array([1.0,  4.2e-05, 0.0])],
            "hitandrun_pure"            : [np.array([1.0, -4.2e-05, 0.0]),
                                            np.array([1.0,  4.2e-05, 0.0])],
            "merge"                      : [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05 ,0.0])],
            "merge_spinner"               : [np.array([1.0, -5.0e-05, 0.0]),
                                            np.array([1.0,  5.0e-05 ,0.0])]
            }

vel_vectors = {"disruption_headon"         : [np.array([ 0.00,  6.280005, 0.0]),
                                            np.array([ 0.00,  3.90,     0.0])],
            "disruption_off_axis"       : [np.array([ 0.00,  6.280005, 0.0]),
                                            np.array([ 0.05,  3.90,     0.0])],
            "supercatastrophic_headon":   [np.array([ 0.00,  6.28,     0.0]),
                                            np.array([ 0.00,  4.28,     0.0])],
            "supercatastrophic_off_axis": [np.array([ 0.00,  6.28,     0.0]),
                                            np.array([ 0.05,  4.28,     0.0])],
            "hitandrun_disrupt"         : [np.array([ 0.00,  6.28,     0.0]),
                                            np.array([-1.45, -6.28,     0.0])],
            "hitandrun_pure"            : [np.array([ 0.00,  6.28,     0.0]),
                                            np.array([-1.52, -6.28,     0.0])],
            "merge"                     : [np.array([ 0.04,  6.28,     0.0]),
                                            np.array([ 0.05,  6.18,     0.0])],
            "merge_spinner"             : [np.array([ 0.04,  6.28,     0.0]),
                                            np.array([ 0.05,  6.18,     0.0])]
            }

rot_vectors = {"disruption_headon"         : [np.array([0.0, 0.0, 1.0e5]),
                                            np.array([0.0, 0.0, -5e5])],
            "disruption_off_axis":        [np.array([0.0, 0.0, 2.0e5]),
                                            np.array([0.0, 0.0, -1.0e5])],
            "supercatastrophic_headon":   [np.array([0.0, 0.0, 1.0e5]),
                                            np.array([0.0, 0.0, -5.0e5])],
            "supercatastrophic_off_axis": [np.array([0.0, 0.0, 1.0e5]),
                                            np.array([0.0, 0.0, -1.0e4])],
            "hitandrun_disrupt"         : [np.array([0.0, 0.0, 0.0]),
                                            np.array([0.0, 0.0, 1.0e5])],
            "hitandrun_pure"            : [np.array([0.0, 0.0, 0.0]),
                                            np.array([0.0, 0.0, 1.0e5])],
            "merge"                     : [np.array([0.0, 0.0, 0.0]),
                                            np.array([0.0, 0.0, 0.0])],
            "merge_spinner"             : [np.array([0.0, 0.0, -1.2e6]),
                                            np.array([0.0, 0.0, 0.0])],
            }

body_Gmass = {"disruption_headon"        : [1e-7, 1e-9],
            "disruption_off_axis"       : [1e-7, 1e-9],
            "supercatastrophic_headon"  : [1e-7, 1e-8],
            "supercatastrophic_off_axis": [1e-7, 1e-8],
            "hitandrun_disrupt"         : [1e-7, 7e-10],
            "hitandrun_pure"            : [1e-7, 7e-10],
            "merge"                     : [1e-7, 1e-8],
            "merge_spinner"             : [1e-7, 1e-8]
            }

tstop = {"disruption_headon"         : 2.0e-3,
        "disruption_off_axis"       : 2.0e-3,
        "supercatastrophic_headon"  : 2.0e-3,
        "supercatastrophic_off_axis": 2.0e-3,
        "hitandrun_disrupt"         : 2.0e-4,
        "hitandrun_pure"            : 2.0e-4,
        "merge"                     : 5.0e-3,
        "merge_spinner"             : 5.0e-3,
        }

nfrag_minimum_expected = {"disruption_headon"         : 20,
                        "disruption_off_axis"       : 20,
                        "supercatastrophic_headon"  : 20,
                        "supercatastrophic_off_axis"  : 20,
                        "hitandrun_disrupt"         : 20,
                        "hitandrun_pure"            : 0,
                        "merge"                     : 0,
                        "merge_spinner"             : 0,
                        }

nfrag_maximum_expected = {"disruption_headon"         : 100,
                        "disruption_off_axis"       : 100,
                        "supercatastrophic_headon"  : 100,
                        "supercatastrophic_off_axis"  : 100,
                        "hitandrun_disrupt"         : 100,
                        "hitandrun_pure"            : 0,
                        "merge"                     : 0,
                        "merge_spinner"             : 0,
                        }

expected_regime = {"disruption_headon"         : "Disruption",
                "disruption_off_axis"       : "Disruption",
                "supercatastrophic_headon"  : "Supercatastrophic disruption",
                "supercatastrophic_off_axis" : "Supercatastrophic disruption",
                "hitandrun_disrupt"         : "Hit and run",
                "hitandrun_pure"            : "Hit and run",
                "merge"                   : "Merge",
                "merge_spinner"           : "Hit and run"
                }
expected_outcome =    {"disruption_headon"         : "calculation converged",
                "disruption_off_axis"       : "calculation converged",
                "supercatastrophic_headon"  : "calculation converged",
                "supercatastrophic_off_axis" : "calculation converged",
                "hitandrun_disrupt"         : "calculation converged",
                "hitandrun_pure"            : "No new fragments generated",
                "merge"                   : "Merging",
                "merge_spinner"           : "No new fragments generated"
                }

density = 3000 * swiftest.AU2M**3 / swiftest.MSun
GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
body_radius = body_Gmass.copy()
for k,v in body_Gmass.items():
    body_radius[k] = [((Gmass/GU)/(4./3.*np.pi*density))**(1./3.) for Gmass in v]

body_radius["hitandrun_disrupt"] = [7e-6, 3.25e-6]
body_radius["hitandrun_pure"] = [7e-6, 3.25e-6]


class TestSwiftestRestart(unittest.TestCase):
    def setUp(self):
            # Initialize a target and surface for testing
            self.tmpdir = tempfile.TemporaryDirectory()
            self.simdir = self.tmpdir.name

            self.tmpdir2 = tempfile.TemporaryDirectory()
            self.simdir_repeat = self.tmpdir2.name

    def tearDown(self):
            # Clean up temporary directory
            self.tmpdir.cleanup()

    def test_restart_collision_snapshot(self):
            """
            Test that a simulation with collisions restarts
            """

            sim = swiftest.Simulation(simdir = self.simdir, integrator = 'symba',
                                dump_cadence = 1, tstep_out = 1 * 365.25,
                                tstop = 365.25 * 1, dt = 0.01,
                                DU = 'km', TU = 'd', MU2KG = 1e15)
            sim.add_solar_system_body(['Mars', 'Phobos', 'Deimos'])

            # Add Mars c_lm from GMM3 (hardcoded so we don't have to rely on pyshtools being installed)
            c_lm = np.array(
                    [[[ 1.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [-8.75021132e-04,  5.90314960e-10, -8.46359039e-05, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [-1.18960349e-05,  3.80429811e-06, -1.59480091e-05, 3.50541857e-05,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 5.12940730e-06,  4.21623326e-06, -9.52539393e-07, 6.45688952e-06,  3.09718541e-07,  0.00000000e+00, 0.00000000e+00],
                    [-1.72668240e-06,  4.83798784e-07, -4.29814806e-06, 3.31265594e-06, -4.64037477e-06, -4.44982150e-06, 0.00000000e+00],
                    [ 1.34648814e-06,  1.80236722e-06,  8.61611560e-07, 9.55649930e-07,  1.00849633e-06,  1.65767871e-06, 2.76208184e-06]],
                    [[ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00, -4.94336174e-11,  4.89346259e-05, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00,  2.51773217e-05,  8.36151763e-06, 2.55745706e-05,  0.00000000e+00,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00,  3.76338480e-06, -8.98081277e-06, -1.93196208e-07, -1.28730468e-05,  0.00000000e+00, 0.00000000e+00],
                    [ 0.00000000e+00,  2.12322092e-06, -1.16588966e-06, 2.71791890e-07, -3.38212383e-06,  3.77983077e-06, 0.00000000e+00],
                    [ 0.00000000e+00, -1.51845907e-06,  1.46912740e-06, 3.33031890e-07,  2.63880041e-06,  1.62290102e-06, 8.21838201e-07]]]
                    )

            sim.modify_body(name = 'Mars', c_lm = c_lm)

            # hardcoded impactors to prevent use of my git repo

            names = np.array(['100', '101', '102', '103', '104', '105', '106', '107', '108',
                    '109', '598'])
            radius = np.array([0.35705834, 0.07421946, 0.14916194, 0.13296731, 0.12915766,
                    0.41544253, 0.46550525, 0.17445848, 0.41888955, 0.45045171,
                    0.49215104 ])
            mass = 10 * np.array([2.86020662e-04, 2.56881727e-06, 2.08522999e-05, 1.47711587e-05,
                    1.35375579e-05, 4.50518599e-04, 6.33802015e-04, 3.33623100e-05,
                    4.61826081e-04, 5.74281193e-04, 7.48988414e-04])
            rh = np.array([[-22124.94631599,  -7600.61943767,  -1953.85576394],
                    [-22099.04261316,  -7594.5053911 ,  -1955.35554083],
                    [-22122.04312455,  -7615.95767634,  -1955.17148458],
                    [-22083.30057286,  -7614.06013637,  -1953.85760781],
                    [-22107.9145528 ,  -7592.30630884,  -1950.69068123],
                    [-22124.83908881,  -7600.49807585,  -1961.07379115],
                    [-22091.70625466,  -7601.39044401,  -1943.88804141],
                    [-22091.96202388,  -7598.98000554,  -1952.81119451],
                    [-22098.41518385,  -7590.14851984,  -1954.86866938],
                    [-22117.17018901,  -7620.37817046,  -1965.44576821],
                    [-22104.19743262,  -7621.76082186,  -1942.37784931]])
            vh =  0.503 * np.array([[  36796.54983816, -110084.06529807,    2987.13829165],
                    [  38083.48622951, -110102.85015476,    2921.36450056],
                    [  36869.30640377, -110484.65220524,    2935.57333344],
                    [  38278.23679524, -110732.77923   ,    2990.69525294],
                    [  37624.67760602, -109538.53678746,    3189.04717925],
                    [  36792.03452032, -109813.24269459,    2613.69041106],
                    [  38540.20902575, -110535.5996834 ,    3651.11554834],
                    [  38768.73186052, -109936.55975385,    3071.4989985 ],
                    [  38141.5822543 , -109706.74112105,    2931.3008305 ],
                    [  36896.44746876, -111001.2491618 ,    2413.7897524 ],
                    [  37671.8566089 , -110761.29152044,    3217.24051514]])

            sim.add_body(name = names, radius = radius, mass = mass, rh = rh, vh = vh)

            try:
                sim.run()
            except Exception as e:
                self.fail(f'Failed initial run with Exception: {e}')

            # restart run

            sim_restart = swiftest.Simulation(simdir = self.simdir, read_data=True, param_file = 'param.restart.in', tstop=365.25 * 10)


            try:
                sim_restart.run()
            except Exception as e:
                self.fail(f'Failed restart with Exception: {e}')

            self.assertTrue(os.path.exists(os.path.join(self.simdir,"param.00000000000000365250.in")))
            self.assertEqual(sim_restart.data.time.size, 11)
            return

    def test_restart_accurate(self):
            '''
            Test that a restarted run gives the same outputs as a full run

            '''
            #Define some common arguments
            run_args = {"tstop": 10, 
                        "dt": 0.01,
                        "tstep_out": 1, 
                        "dump_cadence": 1,  
                        "compute_conservation_values": True}
            
            integrators= ["whm","helio","rmvs","symba"]
            bodies= ['Sun', 'Mercury', 'Venus', 'Earth', 'Mars']
            restart_iout = 500
            param_restart = f'param.{restart_iout:020d}.in'
            for i in integrators: 
                sim = swiftest.Simulation(simdir = self.simdir, integrator = i,**run_args,clean=True)
                sim.add_solar_system_body(bodies)
                try:
                    sim.run()
                except Exception as e:
                    self.fail(f'Failed initial run with Exception: {e}')

                # repeat run with exact same parameters
                sim_repeat = swiftest.Simulation(simdir = self.simdir_repeat, integrator = i,**run_args,clean=True)
                sim_repeat.add_solar_system_body(bodies)

                try:
                    sim_repeat.run()
                except Exception as e:
                    self.fail(f'Failed repeat run with Exception: {e}')

                # Check that the output files are the same
                for var in sim.data.data_vars:
                    if (sim.data[var].dtype.type != np.str_ and np.isnan(sim.data[var].values).any()):
                        idx = np.where(~np.isnan(sim.data[var].values))
                        self.assertTrue((sim.data[var].values[idx] == sim_repeat.data[var].values[idx]).all(), 
                                        f'{i}: {var} values are not equal\n\n{sim.data[var].values[idx]}\n\nREPEAT\n{sim_repeat.data[var].values[idx]}')
                    else:
                        self.assertTrue((sim.data[var].values == sim_repeat.data[var].values).all(), f'{var} values are not equal\n\n{sim.data[var].values}\n\nREPEAT\n{sim_repeat.data[var].values}')

                # restarted run (from halfway mark in this case)
                sim_restart = swiftest.Simulation(simdir = self.simdir,
                                                  read_data = True,
                                                  param_file = param_restart,
                                                  compute_conservation_values = True)

                try:
                    sim_restart.run()
                except Exception as e:
                    self.fail(f'Failed restart run with Exception: {e}')

                # check that the output files are the same
                for var in sim.data.data_vars:
                    if (sim.data[var].dtype.type != np.str_ and np.isnan(sim.data[var].values).any()):
                        idx = np.where(~np.isnan(sim.data[var].values))
                        self.assertTrue((sim.data[var].values[idx] == sim_restart.data[var].values[idx]).all(), 
                                        f'{i}: {var} values are not equal\n\n{sim.data[var].values[idx]}\n\nRESTART\n{sim_restart.data[var].values[idx]}')
                    else:
                        self.assertTrue((sim.data[var].values == sim_restart.data[var].values).all(), 
                                        f'{i}: {var} values are not equal\n\n{sim.data[var].values}\n\nRESTART\n{sim_restart.data[var].values}')
            return


    def test_restart_accurate_collision(self):
            '''
            Test that a restarted run with collisions gives the same outputs as a full run.
            collision outcomes taken from test_fraggle.py

            '''
            seed = [974899978, -2041855733, 14415898, 615945619, 1818808148, 462758063, 762516551, 1276432020]
            
            for style in collision_type:
                sim = swiftest.Simulation(simdir=self.simdir, 
                                          rotation=True, 
                                          compute_conservation_values=True, 
                                          seed = seed,
                                          clean=True)
                sim.add_solar_system_body("Sun")
                sim.add_body(name=names, Gmass=body_Gmass[style], radius=body_radius[style], rh=pos_vectors[style], vh=vel_vectors[style], rot=rot_vectors[style])

                # Set fragmentation parameters
                minimum_fragment_gmass = 0.01 * body_Gmass[style][1] 
                gmtiny = 0.50 * body_Gmass[style][1] 
                sim.set_parameter(encounter_save="both", 
                                gmtiny=gmtiny, 
                                minimum_fragment_gmass=minimum_fragment_gmass, 
                                nfrag_reduction=nfrag_reduction[style])
                try:
                    sim.run(dt=tstop[style]/4, tstop=tstop[style], istep_out=1, dump_cadence=2)
                except Exception as e:
                    self.fail(f'Failed initial run with Exception: {e}')
                
                # repeat run with exact same parameters
                sim_repeat = swiftest.Simulation(simdir=self.simdir_repeat, 
                                                 rotation=True, 
                                                 compute_conservation_values=True, 
                                                 seed = seed,
                                                 clean=True)
                sim_repeat.add_solar_system_body("Sun")
                sim_repeat.add_body(name=names, Gmass=body_Gmass[style], radius=body_radius[style], rh=pos_vectors[style], vh=vel_vectors[style], rot=rot_vectors[style])

                # Set fragmentation parameters
                sim_repeat.set_parameter(encounter_save="both",
                                        gmtiny=gmtiny,
                                        minimum_fragment_gmass=minimum_fragment_gmass,
                                        nfrag_reduction=nfrag_reduction[style])
                try:
                    sim_repeat.run(dt=tstop[style]/4, tstop=tstop[style], istep_out=1, dump_cadence=2)
                except Exception as e:
                    self.fail(f'Failed repeat run with Exception: {e}')

                    # check that the output files are the same

                for var in sim.data.data_vars:
                    if (sim.data[var].dtype.type != np.str_ and np.isnan(sim.data[var].values).any()):
                        idx = np.where(~np.isnan(sim.data[var].values))
                        self.assertTrue((sim.data[var].values[idx] == sim_repeat.data[var].values[idx]).all(), f'{var} values are not equal\n\n{sim.data[var].values[idx]}\n\nREPEAT\n{sim_repeat.data[var].values[idx]}')
                    else:
                        self.assertTrue((sim.data[var].values == sim_repeat.data[var].values).all(), f'{var} values are not equal\n\n{sim.data[var].values}\n\nREPEAT\n{sim_repeat.data[var].values}')

                # restarted run (from the halfway mark in this case)

                restart_iout = 2
                param_restart = f'param.{restart_iout:020d}.in'

                sim.set_parameter(param_file=param_restart)
                try:
                    sim.run()
                except Exception as e:
                    self.fail(f'Failed restart run with Exception: {e}')

                    # check that the output files are the same
                
                for var in sim.data.data_vars:
                    if (sim.data[var].dtype.type != np.str_ and np.isnan(sim.data[var].values).any()):
                        idx = np.where(~np.isnan(sim.data[var].values))
                        self.assertTrue((sim.data[var].values[idx] == sim_repeat.data[var].values[idx]).all(), f'{var} values are not equal\n\n{sim.data[var].values[idx]}\n\nRESTART\n{sim_repeat.data[var].values[idx]}')
                    else:
                        self.assertTrue((sim.data[var].values == sim_repeat.data[var].values).all(), f'{var} values are not equal\n\n{sim.data[var].values}\n\nRESTART\n{sim_repeat.data[var].values}')
                


            return

if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()


#!/usr/bin/env python3
"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""
import swiftest
import unittest
import os
import tempfile
import numpy as np
import warnings
warnings.simplefilter('error', RuntimeWarning)



class TestFraggle(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.tmpdir=tempfile.TemporaryDirectory()
        self.simdir = self.tmpdir.name
        
    def tearDown(self):
        # Clean up temporary directory
        self.tmpdir.cleanup() 
        
    def test_collision_outcomes(self):
        '''
        Check that the head on disruption collision generates fragments and conserves quantities
        '''
        # Colliion initial conditions taken from Fragmentation_Movie.py in the example directory

        collision_type = ["disruption_headon", "disruption_off_axis", "supercatastrophic_headon", "supercatastrophic_off_axis","hitandrun_disrupt", "hitandrun_pure", "merge", "merge_spinner"]

        # ----------------------------------------------------------------------------------------------------------------------
        # To increase the number of bodies generated in each collision type, decrease the value of the corresponding nfrag_reduction number 
        # ----------------------------------------------------------------------------------------------------------------------
        nfrag_reduction = {"disruption_headon" : 10.0,
                "disruption_off_axis"         : 10.0,
                "supercatastrophic_headon"    : 10.0,
                "supercatastrophic_off_axis"  : 10.0,
                "hitandrun_disrupt"           : 10.0,
                "hitandrun_pure"              : 1.0,
                "merge"                       : 1.0,
                "merge_spinner"               : 1.0,
                }


        # These initial conditions were generated by trial and error
        names = ["Target","Projectile"]
        pos_vectors = {"disruption_headon"         : [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05 ,0.0])],
                    "disruption_off_axis"        : [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05 ,0.0])], 
                    "supercatastrophic_headon":   [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05, 0.0])],
                    "supercatastrophic_off_axis": [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05, 0.0])],
                    "hitandrun_disrupt"         : [np.array([1.0, -4.2e-05, 0.0]),
                                                    np.array([1.0,  4.2e-05, 0.0])],
                    "hitandrun_pure"            : [np.array([1.0, -4.2e-05, 0.0]),
                                                    np.array([1.0,  4.2e-05, 0.0])],
                    "merge"                      : [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05 ,0.0])],
                    "merge_spinner"               : [np.array([1.0, -5.0e-05, 0.0]),
                                                    np.array([1.0,  5.0e-05 ,0.0])]                
                    }

        vel_vectors = {"disruption_headon"         : [np.array([ 0.00,  6.280005, 0.0]),
                                                    np.array([ 0.00,  3.90,     0.0])],
                    "disruption_off_axis"       : [np.array([ 0.00,  6.280005, 0.0]),
                                                    np.array([ 0.05,  3.90,     0.0])],
                    "supercatastrophic_headon":   [np.array([ 0.00,  6.28,     0.0]),
                                                    np.array([ 0.00,  4.28,     0.0])],
                    "supercatastrophic_off_axis": [np.array([ 0.00,  6.28,     0.0]),
                                                    np.array([ 0.05,  4.28,     0.0])],
                    "hitandrun_disrupt"         : [np.array([ 0.00,  6.28,     0.0]),
                                                    np.array([-1.45, -6.28,     0.0])],
                    "hitandrun_pure"            : [np.array([ 0.00,  6.28,     0.0]),
                                                    np.array([-1.52, -6.28,     0.0])],
                    "merge"                     : [np.array([ 0.04,  6.28,     0.0]),
                                                    np.array([ 0.05,  6.18,     0.0])],
                    "merge_spinner"             : [np.array([ 0.04,  6.28,     0.0]),
                                                    np.array([ 0.05,  6.18,     0.0])] 
                    }

        rot_vectors = {"disruption_headon"         : [np.array([0.0, 0.0, 1.0e5]),
                                                    np.array([0.0, 0.0, -5e5])],
                    "disruption_off_axis":        [np.array([0.0, 0.0, 2.0e5]),
                                                    np.array([0.0, 0.0, -1.0e5])],
                    "supercatastrophic_headon":   [np.array([0.0, 0.0, 1.0e5]),
                                                    np.array([0.0, 0.0, -5.0e5])],
                    "supercatastrophic_off_axis": [np.array([0.0, 0.0, 1.0e5]),
                                                    np.array([0.0, 0.0, -1.0e4])],
                    "hitandrun_disrupt"         : [np.array([0.0, 0.0, 0.0]),
                                                    np.array([0.0, 0.0, 1.0e5])],
                    "hitandrun_pure"            : [np.array([0.0, 0.0, 0.0]),
                                                    np.array([0.0, 0.0, 1.0e5])],
                    "merge"                     : [np.array([0.0, 0.0, 0.0]),
                                                    np.array([0.0, 0.0, 0.0])],
                    "merge_spinner"             : [np.array([0.0, 0.0, -1.2e6]),
                                                    np.array([0.0, 0.0, 0.0])],
                    }

        body_Gmass = {"disruption_headon"        : [1e-7, 1e-9],
                    "disruption_off_axis"       : [1e-7, 1e-9],
                    "supercatastrophic_headon"  : [1e-7, 1e-8],
                    "supercatastrophic_off_axis": [1e-7, 1e-8],
                    "hitandrun_disrupt"         : [1e-7, 7e-10],
                    "hitandrun_pure"            : [1e-7, 7e-10],
                    "merge"                     : [1e-7, 1e-8],
                    "merge_spinner"             : [1e-7, 1e-8] 
                    }

        tstop = {"disruption_headon"         : 2.0e-3,
                "disruption_off_axis"       : 2.0e-3,
                "supercatastrophic_headon"  : 2.0e-3,
                "supercatastrophic_off_axis": 2.0e-3,
                "hitandrun_disrupt"         : 2.0e-4,
                "hitandrun_pure"            : 2.0e-4,
                "merge"                     : 5.0e-3,
                "merge_spinner"             : 5.0e-3,
                }

        nfrag_minimum_expected = {"disruption_headon"         : 20,
                                "disruption_off_axis"       : 20,
                                "supercatastrophic_headon"  : 20,
                                "supercatastrophic_off_axis"  : 20,
                                "hitandrun_disrupt"         : 20,
                                "hitandrun_pure"            : 0,
                                "merge"                     : 0,
                                "merge_spinner"             : 0,
                                }

        nfrag_maximum_expected = {"disruption_headon"         : 100,
                                "disruption_off_axis"       : 100,
                                "supercatastrophic_headon"  : 100,
                                "supercatastrophic_off_axis"  : 100,
                                "hitandrun_disrupt"         : 100,
                                "hitandrun_pure"            : 0,
                                "merge"                     : 0,
                                "merge_spinner"             : 0,
                                }

        expected_regime = {"disruption_headon"         : "Disruption",
                        "disruption_off_axis"       : "Disruption",
                        "supercatastrophic_headon"  : "Supercatastrophic disruption",
                        "supercatastrophic_off_axis" : "Supercatastrophic disruption",
                        "hitandrun_disrupt"         : "Hit and run",
                        "hitandrun_pure"            : "Hit and run",
                        "merge"                   : "Merge",
                        "merge_spinner"           : "Hit and run"
                        }
        expected_outcome =    {"disruption_headon"         : "calculation converged",
                        "disruption_off_axis"       : "calculation converged",
                        "supercatastrophic_headon"  : "calculation converged",
                        "supercatastrophic_off_axis" : "calculation converged",
                        "hitandrun_disrupt"         : "calculation converged",
                        "hitandrun_pure"            : "No new fragments generated",
                        "merge"                   : "Merging",
                        "merge_spinner"           : "No new fragments generated"
                        }

        density = 3000 * swiftest.AU2M**3 / swiftest.MSun
        GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
        body_radius = body_Gmass.copy()
        for k,v in body_Gmass.items():
            body_radius[k] = [((Gmass/GU)/(4./3.*np.pi*density))**(1./3.) for Gmass in v]

        body_radius["hitandrun_disrupt"] = [7e-6, 3.25e-6] 
        body_radius["hitandrun_pure"] = [7e-6, 3.25e-6] 
        
        with warnings.catch_warnings():
            warnings.simplefilter("error", RuntimeWarning) 
            for style in collision_type:
                sim = swiftest.Simulation(simdir=self.simdir, rotation=True, compute_conservation_values=True)
                sim.add_solar_system_body("Sun")
                sim.add_body(name=names, Gmass=body_Gmass[style], radius=body_radius[style], rh=pos_vectors[style], vh=vel_vectors[style], rot=rot_vectors[style])

                # Set fragmentation parameters
                minimum_fragment_gmass = 0.01 * body_Gmass[style][1] 
                gmtiny = 0.50 * body_Gmass[style][1] 
                sim.set_parameter(collision_model="fraggle", 
                                encounter_save="both", 
                                gmtiny=gmtiny, 
                                minimum_fragment_gmass=minimum_fragment_gmass, 
                                nfrag_reduction=nfrag_reduction[style])
                sim.run(dt=tstop[style]/4, tstop=tstop[style], istep_out=1, dump_cadence=0)
                
                collision_logfile = os.path.join(self.simdir, "collisions.log")
                with open(collision_logfile, "r") as f:
                    content = f.read()
                    
                self.assertIn(expected_outcome[style], content, f'{style}: The collision.log file does not contain "{expected_outcome[style]}"')
            
                regime_name = content.split("Regime:")[1].split('\n')[0].strip() 
                self.assertIn(expected_regime[style], regime_name, f'{style}: The collision.log file does not contain the expected regime name.\nExpected: "{expected_regime[style]}"\nGot: "{regime_name}"')
            
                newbody_count = 0 
                with open(collision_logfile, "r") as f:
                    for line in f:
                        if "Newbody" in line:
                            newbody_count += 1
                            
                self.assertGreaterEqual(newbody_count, nfrag_minimum_expected[style], f"{style}: Expected more than {nfrag_minimum_expected[style]} new bodies, got {newbody_count}") 
                self.assertLessEqual(newbody_count, nfrag_maximum_expected[style], f"{style}: Expected less than {nfrag_maximum_expected[style]} new bodies, got {newbody_count}")
                sim.clean()
        return 


    def test_rotation_direction(self):
        # Tests that the rotation state of a target body is consistent with the impact condition. We test this by checking the
        # sign of the z-axis component of rotation and relative strength of the x and y components relative to the z component for
        # off axis collisions, and also ensure that little rotation occurs for head on collisions.
        
        collision_type = ["disruption_off_axis_clockwise", "disruption_off_axis_counterclockwise", "disruption_headon"]
        nfrag_reduction = 10.0
        names = ["Target","Projectile"]
        pos_vectors =  [np.array([1.0, -5.0e-05, 0.0]),
                        np.array([1.0,  5.0e-05 ,0.0])]
        vel_vectors = {
                    "disruption_headon"                    : [np.array([ 0.00,  6.280005, 0.0]),
                                                              np.array([ 0.00,  3.90,     0.0])],
                    "disruption_off_axis_clockwise"        : [np.array([ 0.00,  6.280005, 0.0]),
                                                              np.array([ 0.05,  3.90,     0.0])], 
                    "disruption_off_axis_counterclockwise" : [np.array([ 0.00,  6.280005, 0.0]),
                                                              np.array([-0.05,  3.90,     0.0])],
                    }

        body_Gmass = np.array([1e-7, 1e-9])
        tstop = 2e-3
        dt = tstop/4
        
        GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
        body_Gmass = np.array([1e-7, 1e-9])
        body_mass = body_Gmass / GU
        density = 3000 * swiftest.AU2M**3 / swiftest.MSun
        body_radius = ((body_mass)/(4./3.*np.pi*density))**(1./3.) 
        
        with warnings.catch_warnings():
            warnings.simplefilter("error", RuntimeWarning) 
            for style in collision_type:        
                sim = swiftest.Simulation(simdir=self.simdir, rotation=True, init_cond_format = "XV", compute_conservation_values=True)
                sim.add_solar_system_body("Sun")
                sim.add_body(name=names, Gmass=body_Gmass, radius=body_radius, rh=pos_vectors, vh=vel_vectors[style])

                # Set fragmentation parameters
                minimum_fragment_gmass = 0.01 * body_Gmass[1] 
                gmtiny = 0.50 * body_Gmass[1] 
                sim.set_parameter(collision_model="fraggle", 
                                encounter_save="both", 
                                gmtiny=gmtiny, 
                                minimum_fragment_gmass=minimum_fragment_gmass, 
                                nfrag_reduction=nfrag_reduction)
                sim.run(dt=dt, tstop=tstop, istep_out=1, dump_cadence=0)        
                
                rot_final = sim.data.isel(time=-1).sel(name='Target').rot
                rotmag_final = rot_final.magnitude().values[()]
                vc = vel_vectors[style][1] - vel_vectors[style][0]
                rc = pos_vectors[1] - pos_vectors[0]
                
                dL = body_mass[1] * np.cross(rc, vc)
                expected_rot = np.rad2deg(dL / (body_mass[0] * body_radius[0]**2 * 0.4))
                rot_error =((rot_final - expected_rot) / rotmag_final).values
                
                # Assert that the absolute value of all components of rot_error is less than 2
                self.assertTrue(np.all(np.abs(rot_error) < 2), f"{style}: The relative error in the final rotation is greater than 2")
                
                
                
                # for the off axis collision styles, assert that the sign of the z component of rotation is consistent with the expected_rot sign
                if style != "disruption_headon":
                    self.assertTrue(np.sign(rot_final.values[2]) == np.sign(expected_rot[2]), f"{style}: The sign of the z component of rotation is not consistent with the expected rotation")
                    
                    # check that the ratio of the z component to either x or y is greater than 100
                    self.assertTrue(np.abs(rot_final.values[2]) > 100*np.abs(rot_final.values[0]), f"{style}: The z component of rotation is not significantly larger than the x component")
                    self.assertTrue(np.abs(rot_final.values[2]) > 100*np.abs(rot_final.values[1]), f"{style}: The z component of rotation is not significantly larger than the y component")
                
                        
if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()


#!/usr/bin/env python3
"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

import swiftest
import unittest
import os
import numpy as np
from numpy.random import default_rng
from astroquery.jplhorizons import Horizons
import datetime
import tempfile

rng = default_rng(seed=123)

major_bodies = ["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"]
param = {}

class TestSwiftestIntegration(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.tmpdir=tempfile.TemporaryDirectory()
        self.simdir = self.tmpdir.name
        
    def tearDown(self):
        # Clean up temporary directory
        self.tmpdir.cleanup() 

    def test_integrators(self):
        """
        Tests that Swiftest is able to integrate a collection of massive bodies and test particles with all available integrators
        """ 
        print("\ntest_integrators: Tests that Swiftest is able to integrate a collection of massive bodies and test particles with all available integrators")
        sim = swiftest.Simulation(simdir=self.simdir)

        # Add the modern planets and the Sun using the JPL Horizons Database.
        sim.add_solar_system_body(major_bodies)
        
        # Add 10 user-defined test particles.
        ntp = 10

        name_tp  = [f"TestParticle_{i:02}" for i in range(1,ntp+1)]
        a_tp     = rng.uniform(0.3, 1.5, ntp)
        e_tp     = rng.uniform(0.0, 0.2, ntp)
        inc_tp   = rng.uniform(0.0, 10, ntp)
        capom_tp = rng.uniform(0.0, 360.0, ntp)
        omega_tp = rng.uniform(0.0, 360.0, ntp)
        capm_tp  = rng.uniform(0.0, 360.0, ntp)

        integrators= ["whm","helio","rmvs","symba"]
        sim.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
        sim.set_parameter(tstart=0.0, tstop=0.02, dt=0.01, istep_out=1, dump_cadence=0)
        integrators= ["whm","helio","rmvs","symba"]
        for i in integrators:
            try:
                sim.run(integrator=i)
            except:
                self.fail(f"Failed with integrator {i}")
        return    
            
    def test_conservation(self):
        """
        Tests that Swiftest conserves mass, energy, and momentum to within acceptable tolerances.
        """
        print("\ntest_conservation: Tests that Swiftest conserves mass, energy, and momentum to within acceptable tolerances.")
        
        # Error limits
        L_slope_limit = 1e-9
        E_slope_limit = 1e-7
        GM_limit = 1e-14
 
        sim = swiftest.Simulation(simdir=self.simdir)
        
        sim.add_solar_system_body(major_bodies)
        
        dt = 0.01
        nout = 100
        tstop = 1e3
        tstep_out = tstop / nout
              
        sim.run(tstart=0.0, tstop=tstop, dt=dt, tstep_out=tstep_out, dump_cadence=0, compute_conservation_values=True, integrator="symba")
        
        def fit_func(x,slope,b):
            """
            Linear function for curve fitting
            """
            return slope * x + b
        
        # Calculate the angular momentum error
        sim.data['L_tot'] = sim.data['L_orbit'] + sim.data['L_rot'] + sim.data['L_escape']
        sim.data['DL'] = sim.data['L_tot'] - sim.data['L_tot'].isel(time=0)
        L_error = sim.data['DL'].magnitude() / sim.data['L_tot'].isel(time=0).magnitude()

        # Calculate the energy error
        E_error = (sim.data['TE'] - sim.data['TE'].isel(time=0)) / sim.data['TE'].isel(time=0)

        # Calculate the mass error
        sim.data['GMtot'] = sim.data['Gmass'].sum(dim='name',skipna=True) + sim.data['GMescape']
        GM_error = (sim.data['GMtot'] - sim.data['GMtot'].isel(time=0)) / sim.data['GMtot'].isel(time=0)
        GM_final = GM_error.isel(time=-1).values
        
        # Compute the slope of the error vs time fit
        E_fit_result = E_error.curvefit("time",fit_func)
        L_fit_result = L_error.curvefit("time",fit_func)
        
        E_slope = E_fit_result['curvefit_coefficients'].sel(param='slope').values
        L_slope = L_fit_result['curvefit_coefficients'].sel(param='slope').values
        
        # Print the final errors
        print("\n")
        print(f"Angular momentum error slope: {L_slope:.2e}/{sim.TU_name}")
        print(f"Energy error slope: {E_slope:.2e}/{sim.TU_name}")
        print(f"Final Mass Error: {GM_final:.2e}")

        self.assertLess(np.abs(L_slope),L_slope_limit, msg=f"Angular Momentum Error of {L_slope:.2e}/{sim.TU_name} higher than threshold value of {L_slope_limit:.2e}/{sim.TU_name}")
        self.assertLess(np.abs(E_slope),E_slope_limit, msg=f"Energy Error of {E_slope:.2e}/{sim.TU_name} higher than threshold value of {E_slope_limit:.2e}/{sim.TU_name}")
        self.assertLess(np.abs(GM_final),GM_limit, msg=f"Mass Error of {GM_final:.2e} higher than threshold value of {GM_limit:.2e}")
        return 
        
    def test_gr(self):
        """
        Tests that GR is working correctly by computing the precession of Mercury's longitude of periapsis and comparing it to
        values obtained from the JPL/Horizons ephemeris service
        """
        print("\ntest_gr: Tests that GR is working correctly.")        
       
        # Error limits 
        dvarpi_limit = 1e-3
        integrators= ["whm","helio","rmvs","symba"] 
        
        # Initialize the simulation object as a variable. Define the directory in which the output will be placed.
        tstep_out = 50
        sim = swiftest.Simulation(simdir=self.simdir, tstop=5000.0, dt=0.01, tstep_out=tstep_out, dump_cadence=0,general_relativity=True)
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

        # Get the start and end date of the simulation so we can compare with the real solar system.
        start_date = sim.ephemeris_date
        tstop_d = float(sim.param['TSTOP'] * sim.TU2S / swiftest.JD2S)

        stop_date = (datetime.datetime.fromisoformat(start_date) + datetime.timedelta(days=tstop_d)).isoformat()

        #Get the ephemerides of Mercury for the same timeframe as the simulation.
        obj = Horizons(id='1', location='@sun',
                    epochs={'start':start_date, 'stop':stop_date,
                            'step':f'{tstep_out}y'})
        el = obj.elements()
        t = (el['datetime_jd']-el['datetime_jd'][0]) / 365.25
        varpi_obs = el['w'] + el['Omega']
        dvarpi_obs = np.diff(varpi_obs) / np.diff(t)
        dvarpi_obs_mean = np.mean(dvarpi_obs) 
        
        for i in integrators:
            sim.run(integrator=i)
            varpi_sim = sim.data['varpi'].sel(name="Mercury")
            dvarpi_gr = np.diff(varpi_sim) / tstep_out
            dvarpi_err = np.mean(dvarpi_obs - dvarpi_gr) / dvarpi_obs_mean
            print(f'{i}: Mercury precession rate error {dvarpi_err:.2e} "/{sim.TU_name}')
            self.assertLess(np.abs(dvarpi_err),dvarpi_limit,msg=f'{dvarpi_err:.2e} /{sim.TU_name} is higher than threshold value of {dvarpi_limit:.2e} "/{sim.TU_name}')

        return
    
         
if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()


#!/usr/bin/env python3
"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

import swiftest
import unittest
import os
import io
import sys
import numpy as np
from numpy.random import default_rng
from astroquery.jplhorizons import Horizons
import datetime
import tempfile
import subprocess

# if python version is >=3.11 use contextlib.chdir, otherwise we need to define it manually 
if sys.version_info >= (3, 11):
    from contextlib import chdir
else:
    from contextlib import contextmanager
    @contextmanager
    def chdir(path):
        oldpwd = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(oldpwd)


rng = default_rng(seed=123)

major_bodies = ["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"]
param = {}

class TestSwiftestIO(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.tmpdir=tempfile.TemporaryDirectory()
        self.simdir = self.tmpdir.name
        
    def tearDown(self):
        # Clean up temporary directory
        self.tmpdir.cleanup() 
        
    def test_jpl_parser(self):
        """
        Tests that the JPL Horizons parser is able to read data and process it correctly for a variety of different cases.
        """
        
        print("\ntest_jpl_parser")
        names = major_bodies + ["Kleopatra","Ceres","Vesta","Pallas","Hygiea","Eris","Dysnomia","Pluto","Charon","Haumea","Quaoar","Orcus","Io","Europa","Titan","Triton","Umbriel","Moon","Phobos","Deimos"]
        test_keys=["mass","radius","rot"]
        
        body_props = {
                    "Sun": {
                        "mass": 1.988409870698051e+30,
                        "radius": 695700000.0,
                        "rot": [
                            2.035154582041381e-05,
                            -7.037127139847693e-05,
                            0.000149334181813172
                        ]
                    },
                    "Mercury": {
                        "mass": 3.3010006367708975e+23,
                        "radius": 2439400.0,
                        "rot": [
                            6.4872186125818495e-06,
                            -5.800032565782477e-06,
                            7.051241078428856e-05
                        ]
                    }, 
                    "Venus": {
                        "mass": 4.867305814842006e+24,
                        "radius": 6051840.0,
                        "rot": [
                            -3.2045928611476063e-07,
                            -1.8640854693792636e-07,
                            -1.7141180404335365e-05
                        ]
                    },
                    "Earth": {
                        "mass": 6.045626290427461e+24,
                        "radius": 6371010.0,
                        "rot": [
                            1.1179382676404839e-05,
                            0.001662031095337799,
                            0.0038332533328792147
                        ]
                    },
                    "Mars": {
                        "mass": 6.416908921385015e+23,
                        "radius": 3389920.0,
                        "rot": [
                            0.0018117998589411537,
                            -0.00022719827850226695,
                            0.003627605423316273
                        ]
                    },
                    "Jupiter": {
                        "mass": 1.8981246258034552e+27,
                        "radius": 69911000.0,
                        "rot": [
                            -0.0001470359129430991,
                            -0.00036060109828205247,
                            0.010067934171122136
                        ]
                    },
                    "Saturn": {
                        "mass": 5.683173701212113e+26,
                        "radius": 58232000.0,
                        "rot": [
                            0.0008024029853603659,
                            0.004339610305746439,
                            0.008281723231138592
                        ]
                    },
                    "Uranus": {
                        "mass": 8.680987153709004e+25,
                        "radius": 25362000.0,
                        "rot": [
                            0.0012296933085501616,
                            0.005614774719295097,
                            -0.0007793665749130923
                        ]
                    },
                    "Neptune": {
                        "mass": 1.024092409690904e+26,
                        "radius": 24624000.0,
                        "rot": [
                            0.0022373803628068065,
                            -0.0019444371136576489,
                            0.005453805355586545
                        ]
                    },
                    "Kleopatra": {
                        "mass": None,
                        "radius": 61000.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.018570102135561744
                        ]
                    },
                    "Ceres": {
                        "mass": 9.383515874323901e+20,
                        "radius": 469700.0,
                        "rot": [
                            0.0015876541706678936,
                            0.00031457177647549044,
                            0.010900790208524158
                        ]
                    },
                    "Vesta": {
                        "mass": 2.5902701406889116e+20,
                        "radius": 262700.0,
                        "rot": [
                            0.008727916415949765,
                            -0.004872733218667549,
                            0.01582674491187599
                        ]
                    },
                    "Pallas": {
                        "mass": 2.0421617248250754e+20,
                        "radius": 256500.0,
                        "rot": [
                            0.010719281540412906,
                            0.006120315206702517,
                            -0.0033835644344280815
                        ]
                    },
                    "Hygiea": {
                        "mass": 1.048799125001873e+20,
                        "radius": 203560.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.007231703789412787
                        ]
                    },
                    "Eris": {
                        "mass": 1.6018159207707176e+22,
                        "radius": 1163000.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.0038610038610038607
                        ]
                    },
                    "Dysnomia": {
                        "mass": 1.4323599478597008e+20,
                        "radius": 350000.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.0
                        ]
                    },
                    "Pluto": {
                        "mass": 1.461465621862967e+22,
                        "radius": 1188300.0,
                        "rot": [
                            -0.0004422702542154748,
                            0.00040738901587759955,
                            -0.00025295769402782636
                        ]
                    },
                    "Charon": {
                        "mass": 1.5896798166099818e+21,
                        "radius": 606000.0,
                        "rot": [
                            -0.00044226841369145176,
                            0.0004073873205131401,
                            -0.0002529566413350254
                        ]
                    },
                    "Haumea": {
                        "mass": 3.96165890055886e+21,
                        "radius": 1050000.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.02554055955790313
                        ]
                    },
                    "Quaoar": {
                        "mass": 1.4323599478597007e+21,
                        "radius": 543500.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.011312985044233772
                        ]
                    },
                    "Orcus": {
                        "mass": 5.777329911751045e+20,
                        "radius": 455000.0,
                        "rot": [
                            0.0,
                            0.0,
                            0.0075826508947528055
                        ]
                    },
                    "Io": {
                        "mass": 8.929648802121572e+22,
                        "radius": 1821490.0,
                        "rot": [
                            -3.4523895639215165e-05,
                            -8.533680868081025e-05,
                            0.0023522483405567646
                        ]
                    },
                    "Europa": {
                        "mass": 4.7985737830184444e+22,
                        "radius": 1560800.0,
                        "rot": [
                            -2.2939528727418718e-05,
                            -3.458615882284886e-05,
                            0.0011729749379113096
                        ]
                    },
                    "Titan": {
                        "mass": 1.3451807680206165e+23,
                        "radius": 2575500.0,
                        "rot": [
                            2.308320022417276e-05,
                            0.00012070667598529457,
                            0.00023060562577272475
                        ]
                    },
                    "Triton": {
                        "mass": 2.1402918658136437e+22,
                        "radius": 1352600.0,
                        "rot": [
                            0.00037532312265244825,
                            -0.0004021490666056507,
                            0.0004472809262462142
                        ]
                    },
                    "Umbriel": {
                        "mass": 1.1720000000000001e+21,
                        "radius": 584700.0,
                        "rot": [
                            -0.000214692403316975,
                            -0.0009726715883856238,
                            0.00013706415384694278
                        ]
                    },
                    "Moon": {
                        "mass": 7.345789170399893e+22,
                        "radius": 1737530.0,
                        "rot": [
                            2.842122641579648e-06,
                            3.0323633838439856e-06,
                            0.00015244753493999662
                        ]
                    },
                    "Phobos": {
                        "mass": 1.08e+16,
                        "radius": 11166.666666666666,
                        "rot": [
                            0.005818491849485793,
                            -0.0004851980251664484,
                            0.011684025938372589
                        ]
                    },
                    "Deimos": {
                        "mass": 1800000000000000.2,
                        "radius": 6300.0,
                        "rot": [
                            0.001331405578408647,
                            -0.0001912658137505407,
                            0.0030123632841272847
                        ]
                    }
                } 
        

        for name in names:
            body = swiftest.init_cond.get_solar_system_body(name=name)
            test_prop = {k: body[k] for k in test_keys if k in body}
            for k,v in body_props[name].items():
                if k in test_keys:
                    if v is None:
                        self.assertIsNone(test_prop[k],msg=f"Error in {name} {k}: {test_prop[k]} != {v}")
                    else:
                        self.assertTrue(np.allclose(test_prop[k],v),msg=f"Error in {name} {k}: {test_prop[k]} != {v}")
        return
       
    def test_xv2el2xv(self):
        """
        Tests that the functions xv2el and el2xv are able to convert between position-velocity and orbital elements without any exceptions being raised
        """
        print("\ntest_xv2el2xv")
        # Generate a set of random position-velocity vectors
        from swiftest.core import xv2el, el2xv
        
        # Test that we can reliably convert between orbital elements and state vectors
        n = 1000
        mu    = np.ones(n)
        a     = rng.uniform(0.1, 1.5, n)
        e     = rng.uniform(0.0, 2.0, n)
        inc   = rng.uniform(0.0, 180.0, n)
        capom = rng.uniform(0.0, 360.0, n)
        omega = rng.uniform(0.0, 360.0, n)
        capm  = rng.uniform(0.0, 360.0, n)
       
        rh, vh = el2xv(mu, a, e, inc, capom, omega, capm)
        a2, e2, inc2, capom2, omega2, capm2, varpi, lam, f, cape, capf = xv2el(mu, rh, vh) 
        
        # Check that the original and converted position-velocity vectors are the same
        self.assertTrue(np.allclose(a,a2),msg=f"Error converting a: {a}, {a2}")
        self.assertTrue(np.allclose(e,e2),msg=f"Error converting a: {e}, {e2}")
        self.assertTrue(np.allclose(inc,inc2),msg=f"Error converting a: {inc}, {inc2}")
        self.assertTrue(np.allclose(capom,capom2),msg=f"Error converting a: {capom}, {capom2}")
        self.assertTrue(np.allclose(omega,omega2),msg=f"Error converting a: {omega}, {omega2}")
        self.assertTrue(np.allclose(capm,capm2),msg=f"Error converting a: {capm}, {capm2}")
        return
    
    def test_gen_ic(self):
        """
        Tests that Swiftest is able to successfully generate a set of initial conditions in a file without any exceptions being raised
        """
        print("\ntest_gen_ic")
        # Files that are expected to be generated:
        file_list = [self.simdir, os.path.join(self.simdir,"param.in"), os.path.join(self.simdir,"init_cond.nc")]
        
        sim = swiftest.Simulation(simdir=self.simdir)

        # Add the modern planets and the Sun using the JPL Horizons Database.
        sim.add_solar_system_body(major_bodies)
        sim.save()
        
        for f in file_list:
            self.assertTrue(os.path.exists(f))

        print("\ntest_read_ic")
        sim2 = swiftest.Simulation(simdir=self.simdir, read_init_cond=True)
        # Add the modern planets and the Sun using the JPL Horizons Database.
        # Check if all names in Dataset read in from file match the expected list of names
        self.assertTrue((major_bodies == sim2.init_cond['name']).all(), msg="Name mismatch in Dataset")
        
        # Check to see if all parameter values read in from file match the expected parameters saved when generating the file
        self.assertTrue(all([v == param[k] for k,v in sim2.param.items() if k in param]))
        return
    
    def test_add_body_combinations(self):
        """
        Tests various combinations of arguments to the Simulation.add_body and Simulation.add_solar_system_body methods to ensure argument checking is working as intended.
        """
        print("\ntest_add_body_combinations")
        sim = swiftest.Simulation(simdir=self.simdir)
        
        # Test that we can can pass a single body:
        sim.add_solar_system_body(name="Sun")
        
        # Test that we can add multiple bodies:
        sim.add_solar_system_body(name=["Mercury","Venus","Earth"])
        
        # Test that we can pass by ephemeris_id alone
        sim.add_solar_system_body(ephemeris_id=["Mars","500"])
      
        # Test that we can initialize a body with only the semimajor axis 
        sim.add_body(a=1.0) 
        self.assertEqual(sim.data.isel(name=-1).e.values[0], 0.0, msg="Failed to initialize body with only semimajor axis")
        self.assertEqual(sim.data.isel(name=-1).inc.values[0], 0.0, msg="Failed to initialize body with only semimajor axis")
        self.assertEqual(sim.data.isel(name=-1).capom.values[0], 0.0, msg="Failed to initialize body with only semimajor axis")
        self.assertEqual(sim.data.isel(name=-1).omega.values[0], 0.0, msg="Failed to initialize body with only semimajor axis")
        self.assertEqual(sim.data.isel(name=-1).capm.values[0], 0.0, msg="Failed to initialize body with only semimajor axis")
        
        # Test that we can input cartesian coordinates
        sim.add_body(mass=1.0, radius=1.0, rh=[1.0,0.0,0.0], vh=[0.0,1.0,0.0])
       
        # orbital elements without semimajor axis 
        with self.assertRaises(ValueError):
           sim.add_body(e=0.0, inc=0.0, capom=0.0, omega=0.0, capm=0.0) 
         
        # Mix of elements and cartesian inputs
        with self.assertRaises(ValueError):
            sim.add_body(a=1.0, e=0.0, rh=[1.0,0.0,0.0], vh=[0.0,1.0,0.0])
            
        # Position but not velocity
        with self.assertRaises(ValueError):
            sim.add_body(rh=[1.0,0.0,0.0])
            
        # Velocity but not position
        with self.assertRaises(ValueError):
            sim.add_body(vh=[0.0,1.0,0.0])
            
        # Add J2 and c_lm values
        with self.assertRaises(ValueError):
            sim.add_body(mass=1.0, radius=1.0, j2rp2=1.0e-6, c_lm=np.ones([2,7]))
            
        # Wrong shape of c_lm
        with self.assertRaises(ValueError):
            sim.add_body(mass=1.0, radius=1.0, c_lm=[1.0,0.0,0.0])
            
        # Mismatched lengths of input arguments
        with self.assertRaises(ValueError):
            sim.add_solar_system_body(name=["Mercury","Venus","Earth"], ephemeris_id=["Mars","500"])
        with self.assertRaises(ValueError):
            sim.add_body(a=[1.0,2.0], e=0.0)
        with self.assertRaises(ValueError):
            sim.add_body(rh=[[1.0,0.0,0.0],[2.0,0.0,0.0]], vh=[0.0,1.0,0.0])     
        with self.assertRaises(ValueError):
            sim.add_body(rh=[1.0,0.0], vh=[0.0,1.0])
      
        # mass and Gmass at the same time 
        with self.assertRaises(ValueError):
            sim.add_body(a=1.0, mass=1.0, radius=1.0, Gmass=4*np.pi**2)
            
        # mass without radius
        with self.assertRaises(ValueError):
            sim.add_body(a=1.0, mass=1.0)
            
        return
    
    def test_mixed_element_cart_input(self):
        """
        Tests that we can mix orbital element and cartesian input
        """
        print("\ntest_mixed_element_cart_input") 
        sim = swiftest.Simulation(simdir=self.simdir)
        
        # Test that we can can pass a single body:
        sim.add_solar_system_body(name="Sun")
        
        sim.add_body(a=[1.0])
        sim.add_body(rh=[2.0,0.0,0.0], vh=[0.0,1.0,0.0])
        try:
            sim.run(tstop=0.02, dt=0.01)
        except:
            self.fail("Failed to run simulation with mixed element and cartesian input")
        
        return
    
    def test_read_multi_dir(self):
        """
        Tests that Swiftest can generate a set of initial conditions, copy them into a new directory, and then run those from a
        different Simulation object (test inspired by Kaustub Anand's workflow)
        """
        print("\ntest_read_multi_dir") 
        def copy_folder_contents(src, dst):
            """
            Copy the contents of the folder 'src' to the folder 'dst'.

            Parameters
            ----------
            src : str
                The path of the source directory.
            dst : str
                The path of the destination directory.

            Returns
            -------
            None
            """
            import shutil

            # Ensure the destination directory exists
            os.makedirs(dst, exist_ok=True)

            # Iterate over all items in the source directory
            for item in os.listdir(src):
                src_path = os.path.join(src, item)
                dst_path = os.path.join(dst, item)

                # Copy files and directories
                if os.path.isdir(src_path):
                    # Recursively copy a directory to a destination
                    shutil.copytree(src_path, dst_path, dirs_exist_ok=True)
                else:
                    # Copy a single file
                    shutil.copy(src_path, dst_path)
            
            return        

        simdir1 = os.path.join(self.simdir, "sim1")
        simdir2 = os.path.join(self.simdir, "sim2")
        sim1 = swiftest.Simulation(simdir=simdir1, tstop=1.0,dt=0.01)
        sim1.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars"])
        sim1.save() 
        copy_folder_contents(simdir1, simdir2)
        sim2 = swiftest.Simulation(read_init_cond=True, simdir=simdir2)
        try:
            sim2.run()
        except:
            self.fail("Failed to run simulation from copied directory")
        
        return

    def test_planetocentric(self):
        """
        Tests that Swiftest is able to set up a simulation in a planetocentric frame and that the results are consistent with the expected values
        """
        print("\ntest_planetocentric")
        sim = swiftest.Simulation(simdir=self.simdir)
        cbname = "Mars"
        cbcenter = "@4"
        satname = ["Phobos","Deimos"]
        sim.add_solar_system_body(cbname)
        sim.add_solar_system_body(satname)

        def get_jpl_data(id):
            tstart = datetime.date.fromisoformat(sim.ephemeris_date)
            tstep = datetime.timedelta(days=1)
            tend = tstart + tstep
            ephemerides_start_date = sim.ephemeris_date
            ephemerides_end_date = tend.isoformat()
            ephemerides_step = '1d' 
            
            jpl = Horizons(id=id, location=cbcenter,
                epochs={'start': ephemerides_start_date, 'stop': ephemerides_end_date,
                'step': ephemerides_step})
            eph = jpl.ephemerides()
            DCONV = swiftest.AU2M / sim.param['DU2M']
            VCONV = (swiftest.AU2M / swiftest.JD2S) / (sim.param['DU2M'] / sim.param['TU2S'])    
        
            vec_tab = jpl.vectors() 
            rx = vec_tab['x'][0] * DCONV
            ry = vec_tab['y'][0] * DCONV
            rz = vec_tab['z'][0] * DCONV
            vx = vec_tab['vx'][0] * VCONV
            vy = vec_tab['vy'][0] * VCONV
            vz = vec_tab['vz'][0] * VCONV
            
            elem_tab = jpl.elements()
            
            a = elem_tab['a'][0] * DCONV
            e = elem_tab['e'][0]
            inc = elem_tab['incl'][0]
            capom = elem_tab['Omega'][0]
            omega = elem_tab['w'][0]
            capm = elem_tab['M'][0]

            rh = np.array([rx,ry,rz]) 
            vh = np.array([vx,vy,vz]) 
            elem = np.array([a,e,inc,capom,omega,capm])
            return rh,vh,elem
                
        for sat in satname:
            sim_rh = sim.data.sel(name=sat,time=0).rh.values
            sim_vh = sim.data.sel(name=sat,time=0).vh.values
            sim_elem = np.array([sim.data[var].sel(name=sat,time=0).values.item() for var in ['a','e','inc','capom','omega','capm']])
            
            jpl_rh,jpl_vh,jpl_elem = get_jpl_data(sat)
            self.assertTrue(np.allclose(sim_rh,jpl_rh),msg=f"Error in rh for {sat}: {sim_rh - jpl_rh}")
            self.assertTrue(np.allclose(sim_vh,jpl_vh),msg=f"Error in vh for {sat}: {sim_vh - jpl_vh}") 
            self.assertTrue(np.allclose(sim_elem,jpl_elem, rtol=1e-4),msg=f"Error in elements for {sat}: {sim_elem - jpl_elem}") 
            
        return
   
    def test_central_body_rotation(self):
        """
        Tests that Swiftest is able to rotate solar system bodies into the equatorial frame of a central body correctly.
        """
        print("\ntest_central_body_rotation")
        sim = swiftest.Simulation(simdir=self.simdir)
        # Precomputed values for [Mars, Phobos, and Deimos] in the ecliptic and Mars equatorial frames
        inc_vals = {"ecliptic": np.array([ np.nan, 26.55406151, 24.06273241]),
                    "mars equator": np.array([ np.nan, 1.0786523 , 2.69218629])}
        rot_vals = {"ecliptic": np.array([[ 57176.05522852,  -7169.83239366, 114478.52090685],
                                          [183617.63838933, -15311.68519899, 368719.81695279],
                                          [ 42015.96468119,  -6035.89004401,  95062.95557518]]),
                    "mars equator": np.array([[-3.08375547e-12, -9.09494702e-13,  1.28163331e+05],
                                              [-3.79488273e+02,  7.76140989e+03,  4.12121104e+05],
                                              [-4.88981793e+03, -1.53941766e+02,  1.03994254e+05]])}
                                                
        # No rotation should keep it in the ecliptic frame 
        sim.add_solar_system_body(["Mars","Phobos","Deimos"]) # Default should be ecliptic
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc,inc_vals["ecliptic"],equal_nan=True)
        rot_close = np.allclose(sim_rot,rot_vals["ecliptic"],rtol=1e-4)
        
        self.assertTrue(inc_close,msg=f"Error in inclination 1")
        self.assertTrue(rot_close,msg=f"Error in rotation 1")

        # Rotating to match the Mars pole
        sim = swiftest.Simulation(simdir=self.simdir)

        sim.add_solar_system_body(["Mars","Phobos","Deimos"],align_to_central_body_rotation=True)
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc,inc_vals["mars equator"],equal_nan=True)
        rot_close = np.allclose(sim_rot,rot_vals["mars equator"])        
       
        self.assertTrue(inc_close,msg=f"Error in inclination 2")
        self.assertTrue(rot_close,msg=f"Error in rotation 2")
        
        sim = swiftest.Simulation(simdir=self.simdir)

        # Mix and match
        sim.add_solar_system_body("Mars") # ecliptic
        sim.add_solar_system_body(["Phobos","Deimos"],align_to_central_body_rotation=True) # mars equator
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc[1:],inc_vals["mars equator"][1:]) 
        rot_close = np.allclose(sim_rot[0:1],rot_vals["ecliptic"][0:1]) and np.allclose(sim_rot[1:],rot_vals["mars equator"][1:]) 
        
        self.assertTrue(inc_close,msg=f"Error in inclination 3")
        self.assertTrue(rot_close,msg=f"Error in rotation 3")

        sim = swiftest.Simulation(simdir=self.simdir)

        sim.add_solar_system_body("Mars",align_to_central_body_rotation=True)
        sim.add_solar_system_body(["Phobos","Deimos"]) # ecliptic 
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc[1:],inc_vals["ecliptic"][1:]) 
        rot_close = np.allclose(sim_rot[0:1],rot_vals["mars equator"][0:1]) and np.allclose(sim_rot[1:],rot_vals["ecliptic"][1:]) 
        
        self.assertTrue(inc_close,msg=f"Error in inclination 4")
        self.assertTrue(rot_close,msg=f"Error in rotation 4")

        sim = swiftest.Simulation(simdir=self.simdir)

        sim.add_solar_system_body("Mars",align_to_central_body_rotation=True)
        sim.add_solar_system_body("Phobos") # ecliptic
        sim.add_solar_system_body("Deimos",align_to_central_body_rotation=True) # mars equator
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc[1:2],inc_vals["ecliptic"][1:2]) and np.allclose(sim_inc[2:],inc_vals["mars equator"][2:])
        rot_close = np.allclose(sim_rot[0:1],rot_vals["mars equator"][0:1]) and np.allclose(sim_rot[1:2],rot_vals["ecliptic"][1:2]) and np.allclose(sim_rot[2:],rot_vals["mars equator"][2:])
        
        self.assertTrue(inc_close,msg=f"Error in inclination 5")
        self.assertTrue(rot_close,msg=f"Error in rotation 5")

        sim = swiftest.Simulation(simdir=self.simdir)

        sim.add_solar_system_body("Mars",align_to_central_body_rotation=True)
        sim.add_solar_system_body("Phobos",align_to_central_body_rotation=True) # mars equator
        sim.add_solar_system_body("Deimos") # ecliptic
        sim_inc = sim.data.isel(time=0).inc.values
        sim_rot = sim.data.isel(time=0).rot.values

        inc_close = np.allclose(sim_inc[1:2],inc_vals["mars equator"][1:2]) and np.allclose(sim_inc[2:],inc_vals["ecliptic"][2:])
        rot_close = np.allclose(sim_rot[0:2],rot_vals["mars equator"][0:2]) and np.allclose(sim_rot[2:2],rot_vals["ecliptic"][2:]) 
        
        self.assertTrue(inc_close,msg=f"Error in inclination 6")
        self.assertTrue(rot_close,msg=f"Error in rotation 6")
        
        return 

    def test_remove_body(self):
        """
        Tests that Swiftest is able to remove a body from the simulation
        """
        print("\ntest_remove_body")
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars"])
        sim.remove_body("Mars")
        self.assertTrue("Mars" not in sim.data.name.values)
        
        with self.assertWarns(Warning) as cm:  # Using Warning as the base class, adjust if needed
            sim.remove_body("Arrakis")
        self.assertIn("Arrakis not found in the Dataset", str(cm.warning))
        
        with self.assertWarns(Warning) as cm:
            sim.remove_body(["Sun", "Mercury","Venus","Earth"])  # Remove all bodies
        self.assertIn("No bodies left in the Dataset", str(cm.warning))
        
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars"])
        sim.remove_body(id=4)
        self.assertTrue("Mars" not in sim.data.name.values)
        with self.assertWarns(Warning) as cm:  # Using Warning as the base class, adjust if needed
            sim.remove_body(id=10)
        self.assertIn("10 not found in the Dataset", str(cm.warning))
        return    
    
    def test_modify_body(self):
        """
        Tests that Swiftest is able to modify the properties of a body in the simulation 
        """
        print("\ntest_modify_body")
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(['Sun','Mercury'])
        sim.modify_body(name='Mercury',a=100.0) 
        self.assertGreater(sim.data.sel(name='Mercury')['a'].values.item(), 99.0)
        
        self.assertTrue('j2rp2' in sim.data)
       
        c_lm = np.ones([2, 3, 3]) 
        sim.modify_body(name='Sun',c_lm = c_lm)
       
        is_close = np.allclose(sim.data.sel(name='Sun')['c_lm'].values,c_lm) 
        self.assertTrue(is_close,msg=f"Error in c_lm: {sim.data.sel(name='Sun')['c_lm'].values - c_lm}")
        
        self.assertTrue('c_lm' in sim.init_cond, msg=f"c_lm not found in sim.init_cond") 
        self.assertTrue('sign' in sim.init_cond.dims, msg=f"sign not found in sim.init_cond.dims")
        self.assertTrue('l' in sim.init_cond.dims, msg=f"l not found in sim.init_cond.dims")
        self.assertTrue('m' in sim.init_cond.dims, msg=f"m not found in sim.init_cond.dims")
        self.assertFalse('j2rp2' in sim.init_cond, msg=f"j2rp2 found in sim.init_cond")
        self.assertFalse('j4rp4' in sim.init_cond, msg=f"j4rp4 found in sim.init_cond")
        self.assertFalse('sign' in sim.init_cond.rh.dims, msg=f"sign found in sim.init_cond.rh.dims")
        self.assertFalse('l' in sim.init_cond.rh.dims, msg=f"l found in sim.init_cond.rh.dims")
        self.assertFalse('m' in sim.init_cond.rh.dims, msg=f"m found in sim.init_cond.rh.dims")
       
        sim.modify_body(name='Sun',j2rp2=1.0e-6,j4rp4=1.0e-8)
        self.assertEqual(sim.data.sel(name='Sun')['j2rp2'].values.item(),1.0e-6)
        self.assertEqual(sim.data.sel(name='Sun')['j4rp4'].values.item(),1.0e-8)
        
        self.assertFalse('c_lm' in sim.init_cond, msg=f"c_lm found in sim.init_cond")
        self.assertFalse('sign' in sim.init_cond.dims, msg=f"sign found in sim.init_cond.dims")
        self.assertFalse('l' in sim.init_cond.dims, msg=f"l found in sim.init_cond.dims")
        self.assertFalse('m' in sim.init_cond.dims, msg=f"m found in sim.init_cond.dims")
        self.assertTrue('j2rp2' in sim.init_cond, msg=f"j2rp2 not found in sim.init_cond")
        self.assertTrue('j4rp4' in sim.init_cond, msg=f"j4rp4 not found in sim.init_cond") 
        
        return
    
    def test_remove_and_modify(self):
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

        # Add 10 user-defined test particles.
        ntp = 10

        name_tp     = ["TestParticle_01", "TestParticle_02", "TestParticle_03", "TestParticle_04", "TestParticle_05", "TestParticle_06", "TestParticle_07", "TestParticle_08", "TestParticle_09", "TestParticle_10"]
        a_tp        = rng.uniform(0.3, 1.5, ntp)
        e_tp        = rng.uniform(0.0, 0.2, ntp)
        inc_tp      = rng.uniform(0.0, 10, ntp)
        capom_tp    = rng.uniform(0.0, 360.0, ntp)
        omega_tp    = rng.uniform(0.0, 360.0, ntp)
        capm_tp     = rng.uniform(0.0, 360.0, ntp)

        sim.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
        sim.modify_body(name="TestParticle_01", a=1.0, e=0.1, inc=0.0, capom=0.0, omega=0.0, capm=0.0)

    def test_read_with_dask(self):
        '''
        Test that a swiftest Simulation data and collisions can be read in with dask
        Adapted from test_fraggle.py

        '''
        # disruption_headon parameters
        names = ["Target", "Projectile"]
        pos_vectors = [np.array([1.0, -5.0e-05, 0.0]), np.array([1.0,  5.0e-05, 0.0])]
        vel_vectors = [np.array([ 0.00,  6.280005, 0.0]), np.array([ 0.00,  3.90, 0.0])]
        rot_vectors = [np.array([0.0, 0.0, 0.0]), np.array([0.0, 0.0, 0.0])]
        body_Gmass =  [1e-7, 1e-9]
        tstop = 2.0e-3
        nfrag_reduction = 10.0
        density = 3000 * swiftest.AU2M**3 / swiftest.MSun
        GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
        body_radius = [((Gmass/GU)/(4./3.*np.pi*density))**(1.0/3.0) for Gmass in body_Gmass]

        sim = swiftest.Simulation(simdir=self.simdir, rotation=True, compute_conservation_values=True)
        sim.add_solar_system_body("Sun")
        sim.add_body(name=names, Gmass=body_Gmass, radius=body_radius, rh=pos_vectors, vh=vel_vectors, rot=rot_vectors)

        # Set fragmentation parameters
        minimum_fragment_gmass = 0.01 * body_Gmass[1] 
        gmtiny = 0.50 * body_Gmass[1] 
        sim.set_parameter(collision_model="fraggle", 
                        encounter_save="both", 
                        gmtiny=gmtiny, 
                        minimum_fragment_gmass=minimum_fragment_gmass, 
                        nfrag_reduction=nfrag_reduction)
        sim.run(dt=tstop/4, tstop=tstop, istep_out=1, dump_cadence=0)

        # read in the data with dask and then save a new set of initial conditions
        try:
            sim_read = swiftest.Simulation(simdir=self.simdir, read_param=True, read_data=True, 
                                            read_collisions=True, dask = True)
            sim_read.save()
        except Exception as e:
            self.fail(f"Failed to read in data with dask: {e}")

        return
   
    def test_newrun_from_old(self):
        '''
        Test that a new set of initial conditions can be extracted from an arbitrary output point of an old run
        '''
        
        run_args={'tstop':1.0,'dt':0.01,'istep_out':1,'dump_cadence':0,'integrator':'whm'} 
        # Build a fresh simulation
        sim = swiftest.Simulation(simdir=self.simdir)
        sim.add_solar_system_body(["Sun"])
        sim.add_body(a=1.0)
        sim.run(**run_args)
       
        # Build a new simulation from the half way point of the old one 
        tmpdir2=tempfile.TemporaryDirectory()
        simdir2 = tmpdir2.name 
        sim.set_parameter(simdir=simdir2)
        sim.save(framenum=50)
        sim2 = swiftest.Simulation(simdir=simdir2, read_init_cond=True)
        tstart = sim2.init_cond.time.values[0]
        sim2.run(tstart=tstart,**run_args)
        
        # Now check if the final states of the two simulations are approximately the same:
        s1=sim.data.isel(name=1,time=np.arange(50,101))
        s2=sim2.data.isel(name=1)
        self.assertTrue(np.allclose(s1.rh.values,s2.rh.values,rtol=1e-12),msg=f"Error in rh: {s1.rh.values - s2.rh.values}")
        self.assertTrue(np.allclose(s1.vh.values,s2.vh.values,rtol=1e-12),msg=f"Error in vh: {s1.vh.values - s2.vh.values}") 
        
        tmpdir2.cleanup()  
        return   

    def test_verbose_flag(self): 
        """
        Tests behavior of the verbose flag that is passed to various functions.
        """   
        # Set up a default system (verbose should be set to True in this case)
        sim = swiftest.Simulation(simdir=self.simdir)
        
        # Capture the output
        captured_output = io.StringIO()
        sys.stdout = captured_output

        # Add the modern planets and the Sun using the JPL Horizons Database.
        sim.add_solar_system_body(["Sun", "Earth"])

        # Assert that something was printed when verbose=True
        self.assertTrue(captured_output.getvalue().strip() != "", "Verbose mode should print output, but nothing was printed.")

        # Reset output capture
        captured_output.truncate(0)
        captured_output.seek(0)

        # Test with verbose turned off
        sim = swiftest.Simulation(clean=True, verbose=False)
        sim.add_solar_system_body(["Sun", "Earth"])
        
        # Assert that nothing was printed when verbose=False
        self.assertEqual(captured_output.getvalue().strip(), "", "Verbose mode is off, but output was printed.")

        # Test that setting the verbose flag using set_parameter sets the verbose attribute of the Simulation object
        sim.set_parameter(verbose=True)
        self.assertTrue(sim.verbose, "Verbose flag should be True after setting it using set_parameter.")

        # Reset output capture
        captured_output.truncate(0)
        captured_output.seek(0)

        # Test that passing a different verbose flag to a function overrides the verbose attribute of the Simulation object
        sim.add_solar_system_body("Mercury", verbose=False)
        
        self.assertTrue(sim.verbose, "Verbose flag should still be set to True after calling a method with verbose=False.")
        sim.run(tstop=0.1,dt=0.01,verbose=False) 
        # Assert that nothing was printed despite verbose=True in the object, because verbose=False was passed to the method
        self.assertEqual(captured_output.getvalue().strip(), "", "Method call with verbose=False should suppress output.")
        
        captured_output.truncate(0)
        captured_output.seek(0)
        
        sim.run(tstart=0,tstop=0.1)
        self.maxDiff = None
        
        self.assertTrue(captured_output.getvalue().strip() != "", "Verbose mode should print output, but nothing was printed.")

        # Clean up by resetting stdout
        sys.stdout = sys.__stdout__
        
    def test_init_cond_format_change(self):
        """
        Test that the user can switch between XV and EL input types even after bodies have been added
        """
        def _run_simulation(start_format, end_format):
            sim = swiftest.Simulation(simdir=self.simdir, init_cond_format=start_format)
            sim.clean(deep=True)
            sim.add_solar_system_body(["Sun", "Jupiter"])
            sim.set_parameter(init_cond_format=end_format,tstop=5, dt=5)
            sim.clean()
            sim.save()
            with chdir(sim.simdir):
                res = subprocess.run(["swiftest","symba","param.in","quiet"], capture_output=True).stdout.decode()
            if "error" in res:
                raise Exception("Error in simulation")
           
        try:
            _run_simulation("XV", "EL")    
        except:
            self.fail("Failed XV->EL")
            
        try:
            _run_simulation("EL", "XV")    
        except:
            self.fail("Failed EL->XV")

    def test_symba_override_options(self):
        # Tests that the `rotation=False` and `compute_conservation_values=False` options are ignored when using the SyMBA integrator, and that
        # a warning will be issued if they are set to False.
        sim = swiftest.Simulation(simdir=self.simdir, integrator="symba")      
        sim.add_solar_system_body(["Sun","Mercury"])   
        with self.assertWarns(UserWarning):
            sim.set_parameter(rotation=False)
        with self.assertWarns(UserWarning):
            sim.set_parameter(compute_conservation_values=False)
        self.assertTrue(sim.param["ROTATION"], "The `rotation` parameter should be set to True when using the SyMBA integrator.")
        self.assertTrue(sim.param["ENERGY"], "The `compute_conservation_values` parameter should be set to True when using the SyMBA integrator.") 
        
        # Test that manually overriding still throws an error on the Fortran side
        sim.param['ENERGY'] = False
        sim.save() 
        with chdir(self.simdir):
            res = subprocess.run(["swiftest","symba","param.in","quiet"], capture_output=True).stdout.decode()
            if "error" not in res:
                self.fail('Failed to throw error when setting ENERGY to "NO" param.in')
                
        sim.param['ENERGY'] = True
        sim.param['ROTATION'] = False
        sim.save()
        with chdir(self.simdir):
            res = subprocess.run(["swiftest","symba","param.in","quiet"], capture_output=True).stdout.decode()
            if "error" not in res:
                self.fail('Failed to throw error when setting ROTATION to "NO" param.in')
        
        
if __name__ == '__main__':
    os.environ["HDF5_USE_FILE_LOCKING"]="FALSE"
    unittest.main()